% TCS Alignment Toolbox
%
% Copyright (C) 2013, 2014
% Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
% AG Theoretical Computer Science
% Centre of Excellence Cognitive Interaction Technology (CITEC)
% University of Bielefeld
% 
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as
% published by the Free Software Foundation, either version 3 of the
% License, or (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU Affero General Public License for more details.
% 
% You should have received a copy of the GNU Affero General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.

% --- Performance Demo ---

%
% In this demonstration we check all possible algorithms for their
% performance on random strings.

%% Preamble
%
% This includes the necessary classes from the TCSAlignmentToolbox
% .jar-file.


clear java;
clear;
clc;

javaaddpath TCSAlignmentToolbox-1.4.jar;
import de.unibi.citec.tcs.alignment.sequence.*;
import de.unibi.citec.tcs.alignment.*;
import de.unibi.citec.tcs.alignment.comparators.*;
import de.unibi.citec.tcs.alignment.wrappers.RandomSequenceGenerator;

%% We first do a performance test for different sequence lengths.
% To do that we calculate the alignment of 2 sequences with some given
% length R times and take the average time. We only 3 keywords of different
% types here.

clear;
clc;

R = 3;

alignSpec = RandomSequenceGenerator.generateSpecification(3);

[ num_algos,  algos, algo_names, lineSpecs, ranges ] = setupAlgorithms(alignSpec);

% now iterate over different lengths.

lengths = 10:10:300;

times = zeros(num_algos,numel(lengths));


for l = 1:numel(lengths);
    L = lengths(l);
    display(sprintf('length=%d',L));
    % generate two sequences with the given length.
    sequences = RandomSequenceGenerator.generateSequences(2,L,alignSpec.getNodeSpecification());
    for algoIdx = 1:num_algos
        display(sprintf('now starting calculation for %s',algo_names{algoIdx}));
        algo = algos(algoIdx);
        % calculate the alignment of both sequences R times.
        for r = 1:R
            tic;
            algo.calculateAlignment(sequences(1),sequences(2));
            times(algoIdx,l) = times(algoIdx,l) + toc;
        end
    end
end

times = times ./ R;

%% plot the results.

close all;

figure(1);

for algoIdx = ranges{1}
    plot(lengths,times(algoIdx,:),lineSpecs{algoIdx});
    hold on;
end

hold off;

title('runtime versus sequence length (global alignment)');
xlabel('sequence length');
ylabel('runtime [s]');
legend(algo_names{ranges{1}});

print_runtime_png('runtime_vs_length_global.png');

figure(2);

for algoIdx = ranges{2}
    plot(lengths,times(algoIdx,:),lineSpecs{algoIdx});
    hold on;
end

hold off;

title('runtime versus sequence length (dynamic time warping)');
xlabel('sequence length');
ylabel('runtime [s]');
legend(algo_names{ranges{2}});

print_runtime_png('runtime_vs_length_dtw.png');

figure(3);

for algoIdx = ranges{3}
    plot(lengths,times(algoIdx,:),lineSpecs{algoIdx});
    hold on;
end

hold off;

title('runtime versus sequence length (local alignment)');
xlabel('sequence length');
ylabel('runtime [s]');
legend(algo_names{ranges{3}});

print_runtime_png('runtime_vs_length_local.png');

%% the next test is on different numbers of sequences in comparing
% iterative or parallel computing. We use 3 keywords here (1 for each
% type).

clear;
clc;

L = 50;
Ns = 5:5:30;

alignSpec = RandomSequenceGenerator.generateSpecification(3);

algo = StrictAlignmentScoreAlgorithm(alignSpec);

times = zeros(2,numel(Ns));

for n = 1:numel(Ns);
    N = Ns(n);
    display(sprintf('number of sequences=%d',N));
    sequences = RandomSequenceGenerator.generateSequences(N,L,alignSpec.getNodeSpecification());
    
    display('iterative processing');
    
    tic;
    for i = 1:N
        for j = 1:N
            algo.calculateAlignment(sequences(i),sequences(j));
        end
    end
    times(1,n) = toc;
    
    display('parallel processing');
    
    parallelEngine = ParallelProcessingEngine(algo,sequences);
    parallelEngine.setNumberOfThreads(8);
    parallelEngine.setVerbose(false);
    tic;
    parallelEngine.calculate();
    times(2,n) = toc;
end


%% plot the results.

close all;

figure(1);

plot(Ns,times(1,:),Ns,times(2,:));

title('runtime versus number of sequences (L=50, K = 30, strict global alignment score)');
xlabel('number of sequences');
ylabel('runtime [s]');
legend('iterative processing','parallel processing');

print_runtime_png('runtime_vs_N.png');

%% We also check how the algorithms scale with the number of
% keywords. We calculate an alignment between two sequences R times and
% take the average runtime.

clear;
clc;

R = 3;
L = 50;
Ks = 3:3:15;

alignSpec = RandomSequenceGenerator.generateSpecification(1);
[ num_algos,  ~, ~, ~, ~ ] = setupAlgorithms(alignSpec);

times = zeros(num_algos,numel(Ks));

% now iterate over different numbers of keywords.

for k = 1:numel(Ks)
    K = Ks(k);
    
    display(sprintf('K=%d',K));
    
    alignSpec = RandomSequenceGenerator.generateSpecification(K);
    [ num_algos,  algos, algo_names, lineSpecs, ranges ] = setupAlgorithms(alignSpec);
    
    % generate two sequences with the given K.
    sequences = RandomSequenceGenerator.generateSequences(2,L,alignSpec.getNodeSpecification());
    for algoIdx = 1:num_algos
        display(sprintf('now starting calculation for %s',algo_names{algoIdx}));
        algo = algos(algoIdx);
        % calculate the alignment of both sequences R times.
        for r = 1:R
            tic;
            algo.calculateAlignment(sequences(1),sequences(2));
            times(algoIdx,k) = times(algoIdx,k) + toc;
        end
    end
end

times = times ./ R;

%% plot the results.

close all;

figure(1);

for algoIdx = ranges{1}
    plot(Ks,times(algoIdx,:),lineSpecs{algoIdx});
    hold on;
end

hold off;

title('runtime versus number of keywords (global alignment)');
xlabel('number of keywords');
ylabel('runtime [s]');
legend(algo_names{ranges{1}});

print_runtime_png('runtime_vs_K_global.png');

figure(2);

for algoIdx = ranges{2}
    plot(Ks,times(algoIdx,:),lineSpecs{algoIdx});
    hold on;
end

hold off;

title('runtime versus number of keywords (dynamic time warping)');
xlabel('number of keywords');
ylabel('runtime [s]');
legend(algo_names{ranges{2}});

print_runtime_png('runtime_vs_K_dtw.png');

figure(3);

for algoIdx = ranges{3}
    plot(Ks,times(algoIdx,:),lineSpecs{algoIdx});
    hold on;
end

hold off;

title('runtime versus number of keywords (local alignment)');
xlabel('number of keywords');
ylabel('runtime [s]');
legend(algo_names{ranges{3}});

print_runtime_png('runtime_vs_K_local.png');

%% We also want to check how well calculation and derivative
% calculation scales with the alphabet size of a symbolic keyword.

clear;
clc;

symbols = java.lang.String('a|b|c|d|e|f|g|h|i|j|k|l|m|n|o|p|q|r|s|t|u|v|w|x|y|z|A|B|C|D|E|F|G|H|I|J|K|L|M|N|O|P|Q|R|S|T|U|V|W|X|Y|Z|');

L = 100;
As = 5:5:50;
R = 3;

% we evaluate four settings: strict alignment, strict derivative, soft
% alignment and soft derivative.
times = zeros(4,numel(As));

for a = 1:numel(As);
    A = As(a);
    
    display(sprintf('A=%d',A));
    
    % create Alphabet
    alpha = Alphabet(symbols.substring(0,2*A));
    % create NodeSpecificaton.
    keywordSpecs = javaArray('de.unibi.citec.tcs.alignment.sequence.KeywordSpecification',1);
    keywordSpecs(1) = SymbolicKeywordSpecification(alpha,'myKeyword');
    nodeSpec = NodeSpecification(keywordSpecs);
    % generate sequences.
    sequences = RandomSequenceGenerator.generateSequences(2,L,nodeSpec);
    % create algorithms.
    alignSpec = AlignmentSpecification(nodeSpec);
    comp = alignSpec.getComparator('myKeyword');

    strict = StrictAlignmentFullAlgorithm(alignSpec);
    soft = SoftAlignmentFullAlgorithm(alignSpec);
    % calculate alignment and derivative.
    for r = 1:R
        % strict alignment
        display('strict global full alignment');
        tic;
        path = strict.calculateAlignment(sequences(1),sequences(2));
        times(1,a) = times(1,a) + toc;
        % strict derivative
        display('strict global full derivative');
        tic;
        path.calculateParameterDerivative(comp,'myKeyword');
        times(2,a) = times(2,a) + toc;
        % soft alignment
        display('soft global full alignment');
        tic;
        pathMat = soft.calculateAlignment(sequences(1),sequences(2));
        times(3,a) = times(3,a) + toc;
        % soft derivative
        display('soft global full derivative');
        tic;
        pathMat.calculateParameterDerivative(comp,'myKeyword');
        times(4,a) = times(4,a) + toc;
    end
end

times = times ./ R;

%% plot the results.

close all;

figure(1);

plot(As,times(1,:),'r-',As,times(2,:),'b-',As,times(3,:),'b-');

title('runtime versus alphabet size (K=1, L=50)');
xlabel('alphabet size');
ylabel('runtime [s]');
legend('strict global full alignment','strict global full derivative','soft global full alignment');

print_runtime_png('runtime_vs_A.png');

figure(2);

plot(As,times(4,:),'b--');

title('soft derivative runtime versus alphabet size (K=1, L=50)');
xlabel('alphabet size');
ylabel('runtime [s]');
legend('soft global full alignment');

print_runtime_png('runtime_vs_A_soft_derivative.png');