/*
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import de.unibi.citec.tcs.alignment.AlignmentPath;
import de.unibi.citec.tcs.alignment.AlignmentSpecification;
import de.unibi.citec.tcs.alignment.StrictAlignmentFullAlgorithm;
import de.unibi.citec.tcs.alignment.comparators.Comparator;
import de.unibi.citec.tcs.alignment.comparators.L1NormComparator;
import de.unibi.citec.tcs.alignment.comparators.TrivialEditComparator;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.sequence.Node;
import de.unibi.citec.tcs.alignment.sequence.Alphabet;
import de.unibi.citec.tcs.alignment.sequence.KeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.NodeSpecification;
import de.unibi.citec.tcs.alignment.sequence.StringKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.SymbolicKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.SymbolicValue;
import de.unibi.citec.tcs.alignment.sequence.StringValue;
import de.unibi.citec.tcs.alignment.sequence.VectorialKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.VectorialValue;
import java.util.Arrays;

/**
 *
 * --- Robot Angles Example ---
 *
 * In this example we compare two artificial sequences of angles of a
 * three-joint robotic arm. This is an example for manual sequence
 * construction and is only recommended for more advanced users. Otherwise
 * please refer to one of the other examples.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class robot_angles {

	public static void main(String[] args) {
		// 1.) Create keyword specifications for three angles, one timestamp,
		// one movement type and a comment.

		final KeywordSpecification[] keywordSpecs = new KeywordSpecification[6];
		keywordSpecs[0] = new VectorialKeywordSpecification(1, "alpha");
		keywordSpecs[1] = new VectorialKeywordSpecification(1, "beta");
		keywordSpecs[2] = new VectorialKeywordSpecification(1, "gamma");
		keywordSpecs[3] = new VectorialKeywordSpecification(1, "timestamp");
		final Alphabet movement_types = new Alphabet("pause|movement");
		keywordSpecs[4] = new SymbolicKeywordSpecification(movement_types, "movement_type");
		keywordSpecs[5] = new StringKeywordSpecification("comment");

		// 2.) Create NodeSpecification.
		final NodeSpecification nodeSpec = new NodeSpecification(keywordSpecs);

		// 3.) Build two sequences for that NodeSpecification.
		final Sequence a = new Sequence(nodeSpec);

		Node n = new Node(a);
		n.setValue("alpha", new VectorialValue(30));
		n.setValue("beta", new VectorialValue(10));
		n.setValue("gamma", new VectorialValue(10));
		n.setValue("timestamp", new VectorialValue(0));
		n.setValue("movement_type", new SymbolicValue(movement_types, "pause"));
		n.setValue("comment", new StringValue(
				"Begin of pause before first movement"));
		a.getNodes().add(n);

		n = new Node(a);
		n.setValue("alpha", new VectorialValue(30));
		n.setValue("beta", new VectorialValue(10));
		n.setValue("gamma", new VectorialValue(10));
		n.setValue("timestamp", new VectorialValue(1));
		n.setValue("movement_type", new SymbolicValue(movement_types, "pause"));
		a.getNodes().add(n);

		n = new Node(a);
		n.setValue("alpha", new VectorialValue(40));
		n.setValue("beta", new VectorialValue(10));
		n.setValue("gamma", new VectorialValue(10));
		n.setValue("timestamp", new VectorialValue(2.2));
		n.setValue("movement_type", new SymbolicValue(movement_types, "movement"));
		n.setValue("comment", new StringValue(
				"Begin of movement. Quite rapidly at the start."));
		a.getNodes().add(n);

		n = new Node(a);
		n.setValue("alpha", new VectorialValue(45));
		n.setValue("beta", new VectorialValue(10));
		n.setValue("gamma", new VectorialValue(10));
		n.setValue("timestamp", new VectorialValue(3));
		n.setValue("movement_type", new SymbolicValue(movement_types, "movement"));
		a.getNodes().add(n);

		n = new Node(a);
		n.setValue("alpha", new VectorialValue(50));
		n.setValue("beta", new VectorialValue(10));
		n.setValue("gamma", new VectorialValue(10));
		n.setValue("timestamp", new VectorialValue(4));
		n.setValue("movement_type", new SymbolicValue(movement_types, "movement"));
		n.setValue("comment", new StringValue(
				"End of movement. End of sequence."));
		a.getNodes().add(n);

		final Sequence b = new Sequence(nodeSpec);

		n = new Node(b);
		n.setValue("alpha", new VectorialValue(30));
		n.setValue("beta", new VectorialValue(10));
		n.setValue("gamma", new VectorialValue(10));
		n.setValue("timestamp", new VectorialValue(0));
		n.setValue("movement_type", new SymbolicValue(movement_types, "pause"));
		n.setValue("comment", new StringValue(
				"Begin of pause before first movement"));
		b.getNodes().add(n);

		n = new Node(b);
		n.setValue("alpha", new VectorialValue(32));
		n.setValue("beta", new VectorialValue(12));
		n.setValue("gamma", new VectorialValue(10));
		n.setValue("timestamp", new VectorialValue(0.8));
		n.setValue("movement_type", new SymbolicValue(movement_types, "movement"));
		n.setValue("comment", new StringValue(
				"Begin of movement."));
		b.getNodes().add(n);

		n = new Node(b);
		n.setValue("alpha", new VectorialValue(36));
		n.setValue("beta", new VectorialValue(7));
		n.setValue("gamma", new VectorialValue(10));
		n.setValue("timestamp", new VectorialValue(2));
		n.setValue("movement_type", new SymbolicValue(movement_types, "movement"));
		b.getNodes().add(n);

		n = new Node(b);
		n.setValue("alpha", new VectorialValue(42));
		n.setValue("beta", new VectorialValue(16));
		n.setValue("gamma", new VectorialValue(10));
		n.setValue("timestamp", new VectorialValue(3.3));
		n.setValue("movement_type", new SymbolicValue(movement_types, "movement"));
		b.getNodes().add(n);

		n = new Node(b);
		n.setValue("alpha", new VectorialValue(50));
		n.setValue("beta", new VectorialValue(10));
		n.setValue("gamma", new VectorialValue(10));
		n.setValue("timestamp", new VectorialValue(4));
		n.setValue("movement_type", new SymbolicValue(movement_types, "movement"));
		n.setValue("comment", new StringValue(
				"End of movement. End of sequence."));
		b.getNodes().add(n);

		// show both sequences.
		System.out.println("Sequence a:");
		System.out.println(a.toString());
		System.out.println("");
		System.out.println("Sequence b:");
		System.out.println(b.toString());
		System.out.println("");

		// 4.) Create alignment specification
		// we want to use all keywords except for the comment keyword.
		// As comparators we use the L1-Norm (abs) for the angles and the time stamp
		// and a TrivialEditComparator for the movement_type.
		// Please have a look in the javadoc for further information on the
		// comparators.
		final String[] align_keywords = new String[5];
		final Comparator[] align_comparators = new Comparator[5];
		for (int k = 0; k < 5; k++) {
			align_keywords[k] = keywordSpecs[k].getKeyword();
			if (k < 4) {
				align_comparators[k] = new L1NormComparator(1);
			} else {
				align_comparators[k] = new TrivialEditComparator();
			}
		}

		// 5.) our weighting should assign emphasis on the alpha angle and the movement
		// type, while the timestamp does not interest us as much.
		final double[] weighting = {0.3, 0.15, 0.15, 0.1, 0.3};

		final AlignmentSpecification alignSpec = new AlignmentSpecification(
				nodeSpec, align_keywords, align_comparators, weighting);

		// visualize the alignment specification with the given defaults
		System.out.println("Alignment Speficiation:");
		System.out.println(alignSpec.toString());
		System.out.println("");

		// 6.) Specify the alignment algorithm
		final StrictAlignmentFullAlgorithm algo = new StrictAlignmentFullAlgorithm(alignSpec);

		// 8.) calculate alignment
		final AlignmentPath alignResult = algo.calculateAlignment(a, b);

		// show the result
		System.out.println("Alignment Result:");
		System.out.println(alignResult.toString());
		System.out.println("");

		// 9.) calculate derivative for the weights. This corresponds to
		// statistics over the distance for the comparators: In the strict case we
		// just sum up over all distances for a given comparator along the optimal
		// alignment path. In our example the gamma angle is consistent for both
		// paths, leading to a derivative of 0. Alpha and Beta have some deviation
		// as do timestamp and movement type. Note that this derivative does _not_
		// consider the current weighting, of course.
		//
		// In a simple learning approach you could now choose to emphasize certain
		// comparators more (or less) if you want to ensure a bigger (lower)
		// distance.
		final double[] weight_derivative = alignResult.calculateWeightDerivative();
		System.out.println("Weight derivative:");
		System.out.println(Arrays.toString(weight_derivative));
		// You could calculate derivatives also for the comparators, but we did not
		// use a derivable comparator in this example.
	}

}
