/*
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import de.unibi.citec.tcs.alignment.AlignmentPath;
import de.unibi.citec.tcs.alignment.AlignmentSpecification;
import de.unibi.citec.tcs.alignment.ParallelDerivativeEngine;
import de.unibi.citec.tcs.alignment.ParallelProcessingEngine;
import de.unibi.citec.tcs.alignment.StrictDTWFullAlgorithm;
import de.unibi.citec.tcs.alignment.StrictDTWScoreAlgorithm;
import de.unibi.citec.tcs.alignment.comparators.EuclideanComparator;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.wrappers.VectorialSequences;
import java.util.Arrays;
import java.util.Random;

/**
 *
 * --- Sine Waves Dynamic Time Warping ---
 *
 * In this example we compare two sine wave sequences using dynamic
 * time warping.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class sine_waves_dtw {

	public static void main(String[] args) {
		// 1.) Create both sine waves. One with doubled frequency.
		final double[][] data = new double[2][];
		final int num_samples = 40;
		data[0] = new double[num_samples / 2];
		data[1] = new double[num_samples];
		for (int i = 0; i < num_samples; i++) {
			if (i < data[0].length) {
				data[0][i] = Math.sin(Math.PI * i * 0.1);
			}
			data[1][i] = Math.sin(Math.PI * i * 0.05);
		}
		//2.) Transform to sequences.
		final Sequence[] sequences = VectorialSequences.toSequences(data);
		//plot one sequence to be sure that the conversion worked.
		System.out.println(sequences[0].toString());
		//3.) Create AlignmentSpecification using a default comparator.
		AlignmentSpecification alignSpec
				= VectorialSequences.setUpSpecification(sequences);
		// show it.
		System.out.println(alignSpec.toString());
		// Maybe we want to use the L2 norm in this case rather than L1.
		alignSpec = VectorialSequences.setUpSpecification(sequences,
				new EuclideanComparator(1));

		//4.) Compare both sequences using Dynamic Time Warping.
		StrictDTWFullAlgorithm dtwPathAlgo
				= new StrictDTWFullAlgorithm(alignSpec);
		AlignmentPath path = dtwPathAlgo.calculateAlignment(sequences[0], sequences[1]);
		//These have fairly low distance because DTW can account for different frequencies.
		System.out.println(path.getScore());

		//5.) Now assume we have two classes of multi-dimensional waves.
		// One class has a slight linear slope in the first dimension, the
		// other one hasn't. That is the discriminative part. However the
		// other dimension is very noisy, which will obscure the discriminative
		// effect under default conditions.
		// create that data set.
		final int numSeqs = 10;
		final int seqLength = 100;
		final Random rand = new Random();
		final double[][][] data2 = new double[numSeqs][][];
		final double slope = 2. / (double) seqLength;
		//shortcut for the class size.
		final int C = numSeqs / 2;
		for (int s = 0; s < numSeqs; s++) {
			data2[s] = new double[seqLength][2];
			//create a nice sine wave in the first dimension with slightly
			//noisy frequencies
			final double f = rand.nextGaussian() * 0.01 + 0.1;
			for (int i = 0; i < seqLength; i++) {
				data2[s][i][0] = Math.sin(Math.PI * i * f);
				//add the slope if we are in class two.
				if (s >= C) {
					data2[s][i][0] += (double) (i - seqLength * 0.5) * slope;
				}
				//put random noise in the second dimension.
				data2[s][i][1] = rand.nextGaussian();
			}
		}
		//6.) give a name to our dimensions and create the sequences.
		final String[] keywords = {"relevant", "noise"};
		final Sequence[] sequences2
				= VectorialSequences.toSequences(data2, keywords);
		System.out.println(sequences2[0]);
		//7.) Create a new AlignmentSpecification
		alignSpec = VectorialSequences.setUpSpecification(sequences2);
		//8.) Let's look at the score matrix.
		StrictDTWScoreAlgorithm dtwScoreAlgo
				= new StrictDTWScoreAlgorithm(alignSpec);
		ParallelProcessingEngine<Double> parallelScoreEngine
				= new ParallelProcessingEngine(dtwScoreAlgo, sequences2);
		parallelScoreEngine.calculate();
		plotScoreMatrix(parallelScoreEngine.getScoreMatrix());
		//9.) This obsiously is not perfect yet.
		// So let's take a closer look at the data space. Using the derivative
		// we can abstract from the current weighting scheme and look which
		// dimension has especially high intra- or inter-class distance. To
		// improve class-discrimination we then can lower the weight on a
		// dimension that as higher (or equal) intra-class distance than inter-
		// class-distance and we can emphasize dimensions with higher inter-
		// class-distance that actually help us to discriminate.
		// this is a very easy learning scheme inspired by the GLVQ cost function.

		//first calculate full paths to have enough information for the derivative..
		dtwPathAlgo = new StrictDTWFullAlgorithm(alignSpec);
		final ParallelProcessingEngine<AlignmentPath> parallelPathEngine
				= new ParallelProcessingEngine(dtwPathAlgo, sequences2);
		parallelPathEngine.calculate();
		final AlignmentPath[][] paths = parallelPathEngine.getResultMatrix();

		//then calculate the weight derivative.
		final ParallelDerivativeEngine derivEngine
				= new ParallelDerivativeEngine(paths);
		final double[][] weightDerivs = derivEngine.calculateWeightDerivatives();

		//then calculate the GLVQ gradient as sketched above.
		final double[] deltaW = new double[2];

		int k = 0;
		for (int i = 0; i < numSeqs; i++) {
			for (int j = 0; j < numSeqs; j++) {
				for (int w = 0; w < 2; w++) {
					if ((i < C && j < C) || (i >= C && j >= C)) {
						// intra class derivatives are added
						deltaW[w] += weightDerivs[k][w];
					} else {
						//inter class derivatives are subtracted
						deltaW[w] -= weightDerivs[k][w];
					}
				}
				k++;
			}
		}

		//show the result.
		System.out.println(Arrays.toString(deltaW));

		// 10.) So we should emphasize the first dimension more than the second
		// one. Let's do that and look at the matrix again.
		alignSpec.setWeighting(new double[]{0.7, 0.3});
		dtwScoreAlgo
				= new StrictDTWScoreAlgorithm(alignSpec);
		parallelScoreEngine
				= new ParallelProcessingEngine(dtwScoreAlgo, sequences2);
		parallelScoreEngine.calculate();
		plotScoreMatrix(parallelScoreEngine.getScoreMatrix());
		// 11.) Better but still not perfect. Let's push it to the extreme.
		alignSpec.setWeighting(new double[]{1, 0.0});
		parallelScoreEngine.calculate();
		plotScoreMatrix(parallelScoreEngine.getScoreMatrix());
		//now we have perfect class separation.
	}

	private static final int DECIMAL_PLACES = 2;

	private static void plotScoreMatrix(final double[][] scoreMat) {
		for (int i = 0; i < scoreMat.length; i++) {
			if (i == 0) {
				//add column labels
				System.out.print("  | ");
				for (int j = 0; j < scoreMat[i].length; j++) {
					System.out.print(j);
					for (int c = 0; c < DECIMAL_PLACES - j / 10; c++) {
						System.out.print(" ");
					}
					System.out.print("| ");
				}
				System.out.println("");
			}
			//add current row label.
			if (i < 10) {
				System.out.print(i + " |");
			} else {
				System.out.print(i + "|");
			}
			for (int j = 0; j < scoreMat[i].length; j++) {
				//add actual content.
				String contentStr = Double.toString(scoreMat[i][j]);
				if (contentStr.length() > DECIMAL_PLACES + 2) {
					contentStr = contentStr.substring(0, 4);
				}
				System.out.print(contentStr);
				//fill up missing spaces if necessary.
				for (int c = contentStr.length(); c < DECIMAL_PLACES + 2; c++) {
					System.out.print(" ");
				}
				System.out.print("|");
			}
			System.out.println("");
		}
	}
}
