/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.comparators.DerivableComparator;
import de.unibi.citec.tcs.alignment.sequence.Value;
import java.util.ArrayList;

/**
 * This is a list-based datastructure accumulating multiple AlignmentPaths. It
 * is also able to calculate the average derivative of all given
 * paths.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class PathList extends ArrayList<AlignmentPath> implements AlignmentDerivativeAlgorithm {

	public PathList() {
		super();
	}

	/**
	 * Returns the average score for all paths in this list.
	 *
	 * @return the average score for all paths in this list.
	 */
	public double getAverageScore() {
		double acumScore = 0;
		for (final AlignmentPath path : this) {
			acumScore += path.getScore();
		}
		return acumScore / (double) this.size();
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public <X extends Value, Y> Y calculateParameterDerivative(
			DerivableComparator<X, Y> comp, String keyword) {
		return comp.transformToResult(calculateRawParameterDerivative(comp, keyword));
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public <X extends Value, Y> double[] calculateRawParameterDerivative(
			DerivableComparator<X, Y> comp, String keyword) {

		final int P = comp.getNumberOfParameters();
		final double[] derivatives = new double[P];

		if (this.isEmpty()) {
			return derivatives;
		}
		/*
		 * now we let each path calculate its derivative and calculate the
		 * average.
		 */
		for (final AlignmentPath path : this) {
			final double[] pathDerivatives = path.calculateRawParameterDerivative(comp, keyword);
			for (int p = 0; p < P; p++) {
				derivatives[p] += pathDerivatives[p];
			}
		}
		for (int p = 0; p < P; p++) {
			derivatives[p] /= (double) this.size();
		}

		return derivatives;
	}

	/**
	 * {@inheritDoc }
	 */
	public double[] calculateWeightDerivative() {
		if (this.isEmpty()) {
			throw new UnsupportedOperationException(
					"Weight derivative without any paths is not defined.");
		}


		/*
		 * now we let each path calculate its weight derivative and take the
		 * average.
		 */
		final int K = get(0).getSpecificaton().size();
		final double[] derivative = new double[K];
		for (final AlignmentPath path : this) {
			final double[] pathDerivative = path.calculateWeightDerivative();
			for (int k = 0; k < K; k++) {
				derivative[k] += pathDerivative[k];
			}
		}
		for (int k = 0; k < K; k++) {
			derivative[k] /= (double) this.size();
		}
		return derivative;
	}

	/**
	 * This transforms the PathList to a matrix, mostly for visualization
	 * purposes. The matrix is equivalent the dynamic programming algorithm of
	 * the StrictAlgorithm that created this PathMap. However, only those
	 * entries are non-zero that are relevant for at least one path in this map.
	 * The value of the output matrix is defined as the average over
	 * all path toMatrix outputs.
	 *
	 * @return the dynamic programming matrix of the algorithm that created this
	 * PathList with non-zero entries at positions that are relevant for at
	 * least
	 * one path.
	 */
	public double[][] toMatrix() {

		final int m = get(0).getLeft().getNodes().size();
		final int n = get(0).getRight().getNodes().size();

		final double[][] out = new double[m + 1][n + 1];
		for (final AlignmentPath path : this) {
			int i = 0;
			int j = 0;
			double oldScore = 0;
			for (final Operation op : path.getOperations()) {
				switch (op.getType()) {
					case DELETION:
					case SKIPDELETION:
						i++;
						break;
					case INSERTION:
					case SKIPINSERTION:
						j++;
						break;
					case REPLACEMENT:
						i++;
						j++;
						break;
					default:
						throw new UnsupportedOperationException("Unsupported operation: " + op.getType());
				}
				out[i][j] += oldScore + op.getWeightedLocalCost();
				oldScore = out[i][j];
			}
		}

		for (int i = 0; i < m; i++) {

			for (int j = 0; j < m; j++) {
				if (out[i][j] > 0) {
					out[i][j] /= (double) this.size();
				}
			}
		}
		return out;
	}
}
