/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.sequence.Sequence;

/**
 * This calculates the soft (global) alignment of two sequences similarly to the
 * Needleman-Wunsch-Algorithm. The whole trick of this
 * SoftAlignmentScoreAlgorithm-Algorithm is to replace the strict minimum in the
 * cost function by a softmin approximation. Thereby we give the alignment a
 * "probabilistic touch": Even suboptimal alignments are considered by the
 * softmin approach, but less than the optimal alignment. More information about
 * softmin can be found in the Softmin class.
 *
 * This algorithm returns just the accumulated, normalized score. In each step
 * the score is 1 at max and in the worst case m+n steps are needed. Thus the
 * normalization is done by dividing by m+n.
 *
 * Normally it is not useful to apply this algorithm. Soft alignment is only
 * used if you want to calculate a derivative.
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class SoftAlignmentScoreAlgorithm
		extends AbstractGapAlignmentAlgorithm<SoftAlignmentScoreAlgorithm.ScoreEntry, Double> {

	public static final double DEFAULTBETA = 1;
	private double beta = DEFAULTBETA;
	private double approxThreshold = Softmin.DEFAULTAPPROXTHRESHOLD;

	/**
	 * This sets up an AlignmentAlgorithm instance according to the given
	 * specification.
	 *
	 * @param alignmentSpecification an AlignmentSpecification.
	 */
	public SoftAlignmentScoreAlgorithm(AlignmentSpecification alignmentSpecification) {
		super(alignmentSpecification, SoftAlignmentScoreAlgorithm.ScoreEntry.class, Double.class);
	}

	/**
	 *
	 * @param beta The parameter defining the "softness" of the alignment. For
	 * beta towards infinity this alignment becomes closer to the strict
	 * alignment. For beta = 0 all possible alignments are equally considered
	 * and softmin returns the average. Please note that a low beta value might
	 * lead to a very rough approximation and that for higher sequence lengths
	 * beta has to be higher, too.
	 */
	public void setBeta(double beta) {
		this.beta = beta;
	}

	/**
	 *
	 * @return The parameter defining the "softness" of the alignment. For beta
	 * towards infinity this alignment becomes closer to the strict alignment.
	 * For beta = 0 all possible alignments are equally considered and softmin
	 * returns the average. Please note that a low beta value might lead to a
	 * very rough approximation and that for higher sequence lengths beta has to
	 * be higher, too.
	 */
	public double getBeta() {
		return beta;
	}

	/**
	 * For more information on the ApproxThreshold have a look at the
	 * Softmin class.
	 *
	 * @return
	 */
	public double getApproxThreshold() {
		return approxThreshold;
	}

	/**
	 * For more information on the ApproxThreshold have a look at the
	 * Softmin class.
	 *
	 * @param approxThreshold
	 */
	public void setApproxThreshold(double approxThreshold) {
		this.approxThreshold = approxThreshold;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public ScoreEntry createInitial() {
		return new ScoreEntry(0);
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public ScoreEntry createDelInitial(ScoreEntry delOld, int i, double delLocal) {
		return new ScoreEntry(delOld.getScore() + delLocal);
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public ScoreEntry createInsInitial(ScoreEntry insOld, int j, double insLocal) {
		return new ScoreEntry(insOld.getScore() + insLocal);
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public ScoreEntry createNewEntry(ScoreEntry delOld, ScoreEntry insOld,
			ScoreEntry repOld,
			int i, int j,
			double delLocal, double insLocal, double repLocal) {
		final double[] choices = new double[3];
		choices[0] = delOld.getScore() + delLocal;
		choices[1] = insOld.getScore() + insLocal;
		choices[2] = repOld.getScore() + repLocal;
		return new ScoreEntry(Softmin.softmin(beta, approxThreshold, choices));
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public Double transformToResult(ScoreEntry[][] alignmentMatrix,
			Sequence a, Sequence b) {
		final double rawScore = alignmentMatrix[a.getNodes().size()][b.getNodes().size()].
				getScore();
		/*
		 * In each step of the alignment we had a distance of 1 at most. Thus we
		 * can normalize the score to the range [0,1] if we just divide the raw
		 * score by the worst case, which would be to delete a entirely and
		 * insert b entirely => worst case score = m+n.
		 */
		return rawScore / ((double) (a.getNodes().size() + b.getNodes().size()));
	}

	public static class ScoreEntry implements AlignmentMatrixEntry {

		private final double totalScore;

		public ScoreEntry(double totalScore) {
			this.totalScore = totalScore;
		}

		public double getScore() {
			return totalScore;
		}
	}
}
