/*
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.sequence.Sequence;
import java.util.EnumMap;

/**
 * This implements a local affine alignment similar to Smith-Waterman
 * and Gotoh. More information about this alignment can be found in the
 * AbstractLocalAlignmentAlgorithm.
 *
 * Note that you are allowed to specify higher deletion, insertion or comparison
 * costs than 1 in your comparators here, because it is still possible to skip
 * both sequences entirely in the worst case. It is, however, still recommended
 * to set the local comparator costs between 0 and 1 as well.
 *
 * The end result is normalized by the worst possible alignment, meaning by the
 * costs to skip both input sequences entirely.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class StrictLocalAlignmentScoreAlgorithm
		extends AbstractLocalAlignmentAlgorithm<Double> {

	public StrictLocalAlignmentScoreAlgorithm(
			AlignmentSpecification alignmentSpecification) {
		super(Double.class, alignmentSpecification);
	}

	/**
	 * This implements the strict maximum of the given input doubles.
	 *
	 * @return the strict maximum of the given input doubles.
	 */
	@Override
	public double choice(double... choices) {
		if (choices == null || choices.length == 0) {
			throw new UnsupportedOperationException("Choice input was invalid!");
		}
		double min = choices[0];
		for (int i = 1; i < choices.length; i++) {
			if (choices[i] < min) {
				min = choices[i];
			}
		}
		return min;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public Double transformToResult(EnumMap<Recurrence, double[][]> dp_tables,
			double[][] compareMatrix,
			double[] deletionMatrix, double[] insertionMatrix,
			double[] skipDeletionMatrix, double[] skipInsertionMatrix,
			Sequence a, Sequence b) {
		final double accum_score = dp_tables.get(Recurrence.SKIPDEL_START)[0][0];
		final double score;
		if (accum_score > 0) {
			//normalize it by the cost for skipping both sequences entirely.
			double worstScore = 0;
			for (int i = 0; i < skipDeletionMatrix.length; i++) {
				worstScore += skipDeletionMatrix[i];
			}
			for (int j = 0; j < skipInsertionMatrix.length; j++) {
				worstScore += skipInsertionMatrix[j];
			}
			score = Math.min(1, accum_score / worstScore);
		} else {
			score = 0;
		}
		return score;
	}

}
