/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment.comparators;

import de.unibi.citec.tcs.alignment.sequence.ValueType;
import de.unibi.citec.tcs.alignment.sequence.VectorialValue;

/**
 * Compares vectors using the L2-Norm. Unfortunately the norm is not normalized
 * between 0 and 1, thus a normalizer is required.
 *
 * Furthermore it is not trivial to include gaps and skips. The comparison
 * counterpart is unclear in that case. We allow two mechanisms in this case:
 * <ol><li>To compare with a fixed vector (the 0-vector by default)</li>
 * <li>To just return a constant term</li></ol>
 *
 * Both are implemented in the ComparisonBasedSkipExtendedComparator
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class EuclideanComparator
		extends ComparisonBasedSkipExtendedComparator<VectorialValue> {

	private Normalizer normalizer;

	public EuclideanComparator(int dimensions) {
		this(new HyperbolicNormalizer(), dimensions);
	}

	public EuclideanComparator(Normalizer normalizer, int dimensions) {
		this(normalizer, dimensions, 0);
	}

	public EuclideanComparator(int dimensions, double comparisonValue) {
		this(new HyperbolicNormalizer(), dimensions, comparisonValue);
	}

	public EuclideanComparator(Normalizer normalizer, int dimensions, double comparisonValue) {
		this(normalizer, getConstantVector(dimensions, comparisonValue),
				getConstantVector(dimensions, comparisonValue),
				getConstantVector(dimensions, comparisonValue),
				getConstantVector(dimensions, comparisonValue));
	}

	public EuclideanComparator(
			VectorialValue deleteComparison,
			VectorialValue insertComparison,
			VectorialValue skipDeleteComparison,
			VectorialValue skipInsertComparison) {
		this(new HyperbolicNormalizer(),
				deleteComparison, insertComparison,
				skipDeleteComparison, skipInsertComparison);
	}

	public EuclideanComparator(Normalizer normalizer,
			VectorialValue deleteComparison,
			VectorialValue insertComparison,
			VectorialValue skipDeleteComparison,
			VectorialValue skipInsertComparison) {
		super(deleteComparison, insertComparison, skipDeleteComparison, skipInsertComparison);
		this.normalizer = normalizer;
	}

	private static VectorialValue getConstantVector(int dimensions, double value) {
		final double[] vec = new double[dimensions];
		if (value != 0) {
			for (int d = 0; d < dimensions; d++) {
				vec[d] = value;
			}
		}
		return new VectorialValue(vec);
	}

	/**
	 * The euclidian distance per default is bound by 0 but has no upper bound.
	 * A normalizer ensures that the result stays in the interval [0,1].
	 *
	 * This is the HyperbolicNormalizer per default.
	 *
	 * @param normalizer a Normalizer.
	 */
	public void setNormalizer(Normalizer normalizer) {
		this.normalizer = normalizer;
	}

	/**
	 * Please refer to the corresponding setter method for more information.
	 *
	 * @return a Normalizer.
	 */
	public Normalizer getNormalizer() {
		return normalizer;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double compare(VectorialValue a, VectorialValue b) {
		if (a == null && b == null) {
			return 0;
		}
		if (a == null && b != null) {
			return insert(b);
		}
		if (a != null && b == null) {
			return delete(a);
		}
		return normalizer.normalize(distance(a.getVector(), b.getVector()));
	}

	/**
	 * Returns the euclidian distance between the vectors a and b, given that
	 * they have the same length.
	 *
	 * @param a a double vector.
	 * @param b a double vector.
	 * @return the euclidian distance between the vectors a and b.
	 */
	public double distance(double[] a, double[] b) {
		double d = 0;
		double localdist;
		for (int i = 0; i < a.length; i++) {
			localdist = a[i] - b[i];
			d += localdist * localdist;
		}
		return Math.sqrt(d);
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public ValueType getType() {
		return ValueType.VECTOR;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public int hashCode() {
		int hash = 7;
		hash = 47 * hash + (this.normalizer != null ? this.normalizer.hashCode() : 0);
		hash = 47 * hash + super.hashCode();
		return hash;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final EuclideanComparator other = (EuclideanComparator) obj;
		if (this.normalizer != other.normalizer
				&& (this.normalizer == null || !this.normalizer.equals(other.normalizer))) {
			return false;
		}
		if (!super.equals(obj)) {
			return false;
		}
		return true;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double comparisonDerivative(int paramIdx, VectorialValue a, VectorialValue b) {
		//this has no further parameters.
		return 0;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public SparseLocalDerivative sparseComparisonDerivative(VectorialValue a, VectorialValue b) {
		//this has no further parameters.
		return new SparseLocalDerivative.EmptySparseLocalEntry();
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public int getNumberOfComparisonParameters() {
		//this has no further parameters.
		return 0;
	}
}
