/*
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment.comparators;

import de.unibi.citec.tcs.alignment.OperationType;
import de.unibi.citec.tcs.alignment.sequence.Value;

/**
 * This is an abstract class to be used for Comparators that are only
 * to be used for comparisons per se but can be extended for use with
 * gaps and skips.
 *
 * This provides a default extension with constant gap and skip costs.
 *
 * It also provides an implementation to make this comparator derivable for
 * the constant gap and skip costs.
 *
 * The derivative of this Comparator is a vector containing at least the
 * derivatives for
 *
 * <ol><li>deletion cost (1 per default)</li>
 * <li>insertion cost (1 per default)</li>
 * <li>skip deletion cost (0.5 per default)</li>
 * <li>skip insertion cost (0.5 per default)</li></ol>
 *
 * additional parameters might be specified by implementing classes.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 * @param <V> the value type this Comparator is defined on.
 */
public abstract class SkipExtendedComparator<V extends Value> implements SkipComparator<V>,
		SparseDerivableComparator<V, double[]> {

	private double deletionCost = 1;
	private double insertionCost = 1;
	private double skipDeletionCost = 0.5;
	private double skipInsertionCost = 0.5;

	public SkipExtendedComparator() {
	}

	/**
	 * Sets the (constant) cost for deleting a value.
	 *
	 * @param deletionCost the new (constant) cost for deleting a value.
	 */
	public void setDeletionCost(double deletionCost) {
		this.deletionCost = deletionCost;
	}

	/**
	 * Returns the (constant) cost for deleting a value.
	 *
	 * @return the (constant) cost for deleting a value.
	 */
	public double getDeletionCost() {
		return deletionCost;
	}

	/**
	 * Sets the (constant) cost for inserting a value.
	 *
	 * @param insertionCost the new (constant) cost for inserting a value.
	 */
	public void setInsertionCost(double insertionCost) {
		this.insertionCost = insertionCost;
	}

	/**
	 * Returns the (constant) cost for inserting a value.
	 *
	 * @return the (constant) cost for inserting a value.
	 */
	public double getInsertionCost() {
		return insertionCost;
	}

	/**
	 * Sets the (constant) cost for deleting a value using a skip.
	 *
	 * @param skipDeletionCost the new (constant) cost for deleting a value
	 * using a skip.
	 */
	public void setSkipDeletionCost(double skipDeletionCost) {
		this.skipDeletionCost = skipDeletionCost;
	}

	/**
	 * Returns the (constant) cost for deleting a value using a skip.
	 *
	 * @return the (constant) cost for deleting a value using a skip.
	 */
	public double getSkipDeletionCost() {
		return skipDeletionCost;
	}

	/**
	 * Sets the (constant) cost for inserting a value using a skip.
	 *
	 * @param skipInsertionCost the new (constant) cost for inserting a value
	 * using a skip.
	 */
	public void setSkipInsertionCost(double skipInsertionCost) {
		this.skipInsertionCost = skipInsertionCost;
	}

	/**
	 * Returns the (constant) cost for inserting a value using a skip.
	 *
	 * @return the (constant) cost for inserting a value using a skip.
	 */
	public double getSkipInsertionCost() {
		return skipInsertionCost;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double delete(V a) {
		return deletionCost;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double insert(V b) {
		return insertionCost;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double skipDelete(V a) {
		return skipDeletionCost;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double skipInsert(V b) {
		return skipInsertionCost;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double calculateLocalDerivative(int paramIdx, V a, V b, OperationType type) {
		if (paramIdx > 3) {
			if (type == OperationType.REPLACEMENT) {
				return comparisonDerivative(paramIdx, a, b);
			} else {
				return 0;
			}
		}
		if (paramIdx == 0 && type == OperationType.DELETION) {
			return 1;
		}
		if (paramIdx == 1 && type == OperationType.INSERTION) {
			return 1;
		}
		if (paramIdx == 2 && type == OperationType.SKIPDELETION) {
			return 1;
		}
		if (paramIdx == 3 && type == OperationType.SKIPINSERTION) {
			return 1;
		}
		return 0;
	}

	/**
	 * This should implement the local derivative calculation in case of
	 * comparisons. If no parameters exist here, it can just return 0.
	 *
	 * @param paramIdx the index of the parameter the alignment is currently
	 * derived to. Note that the first four parameters (0-3) are reserved for
	 * the parameters of this superclass.
	 * @param a the value on the left side of the comparison.
	 * @param b the value on the right side of the comparison.
	 * @return the local derivative for this comparison.
	 */
	public abstract double comparisonDerivative(int paramIdx, V a, V b);

	/**
	 * {@inheritDoc }
	 */
	@Override
	public SparseLocalDerivative calculateSparseLocalDerivative(V a, V b, OperationType type) {
		final SparseLocalDerivative.TrivialSparseLocalDerivative out
				= new SparseLocalDerivative.TrivialSparseLocalDerivative();
		final int paramIdx;
		switch (type) {
			case DELETION:
				paramIdx = 0;
				break;
			case INSERTION:
				paramIdx = 1;
				break;
			case SKIPDELETION:
				paramIdx = 2;
				break;
			case SKIPINSERTION:
				paramIdx = 3;
				break;
			case REPLACEMENT:
				return sparseComparisonDerivative(a, b);
			default:
				throw new UnsupportedOperationException("Unknown operation type: " + type);
		}
		out.add(paramIdx, 1);
		return out;
	}

	/**
	 * This should implement the local derivative calculation in case of
	 * comparisons. If no parameters exist here, it can just return 0.
	 * Note that the first four parameters (0-3) are reserved for
	 * the parameters of this superclass.
	 *
	 * @param a the value on the left side of the comparison.
	 * @param b the value on the right side of the comparison.
	 * @return the local derivative for this comparison.
	 */
	public abstract SparseLocalDerivative sparseComparisonDerivative(V a, V b);

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double[] transformToResult(double[] derivatives) {
		return derivatives;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public int getNumberOfParameters() {
		return 4 + getNumberOfComparisonParameters();
	}

	/**
	 * Implementing classes should return the number of (derivative-relevant)
	 * parameters used for the compare-function.
	 *
	 * @return the number of (derivative-relevant)
	 * parameters used for the compare-function.
	 */
	public abstract int getNumberOfComparisonParameters();

	/**
	 * {@inheritDoc }
	 */
	@Override
	public int hashCode() {
		int hash = 7;
		hash = 97 * hash + (int) (Double.doubleToLongBits(this.deletionCost) ^ (Double.doubleToLongBits(this.deletionCost) >>> 32));
		hash = 97 * hash + (int) (Double.doubleToLongBits(this.insertionCost) ^ (Double.doubleToLongBits(this.insertionCost) >>> 32));
		hash = 97 * hash + (int) (Double.doubleToLongBits(this.skipDeletionCost) ^ (Double.doubleToLongBits(this.skipDeletionCost) >>> 32));
		hash = 97 * hash + (int) (Double.doubleToLongBits(this.skipInsertionCost) ^ (Double.doubleToLongBits(this.skipInsertionCost) >>> 32));
		return hash;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final SkipExtendedComparator<?> other = (SkipExtendedComparator<?>) obj;
		if (Double.doubleToLongBits(this.deletionCost) != Double.doubleToLongBits(other.deletionCost)) {
			return false;
		}
		if (Double.doubleToLongBits(this.insertionCost) != Double.doubleToLongBits(other.insertionCost)) {
			return false;
		}
		if (Double.doubleToLongBits(this.skipDeletionCost) != Double.doubleToLongBits(other.skipDeletionCost)) {
			return false;
		}
		if (Double.doubleToLongBits(this.skipInsertionCost) != Double.doubleToLongBits(other.skipInsertionCost)) {
			return false;
		}
		return true;
	}

	@Override
	public Class<double[]> getResultClass() {
		return double[].class;
	}

}
