/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment.comparators;

import java.util.Iterator;
import java.util.Map;
import java.util.Stack;

/**
 * This is an interface for classes that are able to hold sparse representations
 * of a local derivative, where only for one or a few parameters the derivative
 * is non-zero. This is actually an extension of the java.util.Map-interface for
 * an Integer to Double-Map, where the integer key is the parameter index and
 * the double value is the local derivative for the parameter with that index.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public interface SparseLocalDerivative
		extends Iterable<SparseLocalDerivative.SparseDeriativeEntry> {

	/**
	 * This adds a new derivative to this sparse representation.
	 *
	 * @param paramIdx the index of the parameter for which we have a non-zero
	 * derivative.
	 * @param derivative the derivative value itself.
	 */
	public void add(int paramIdx, double derivative);

	public static class FlexibleSparseLocalDerivative extends Stack<SparseDeriativeEntry>
			implements SparseLocalDerivative {

		public FlexibleSparseLocalDerivative() {
		}

		/**
		 * {@inheritDoc }
		 */
		@Override
		public void add(int paramIdx, double derivative) {
			super.add(new SparseDeriativeEntry(paramIdx, derivative));
		}
	}

	/**
	 * This implements the SparseLocalDerivative interface for a fixed number of
	 * entries.
	 */
	public static class FixedSparseLocalDerivative implements SparseLocalDerivative,
			Iterator<SparseDeriativeEntry> {

		private final int[] indices;
		private final double[] derivatives;
		private int fillIdx = 0;
		private int itIdx = 0;

		public FixedSparseLocalDerivative(int size) {
			this.indices = new int[size];
			this.derivatives = new double[size];
		}

		/**
		 * {@inheritDoc }
		 */
		@Override
		public Iterator<SparseDeriativeEntry> iterator() {
			itIdx = fillIdx;
			return this;
		}

		/**
		 * {@inheritDoc }
		 */
		@Override
		public boolean hasNext() {
			return itIdx > 0;
		}

		/**
		 * {@inheritDoc }
		 */
		@Override
		public SparseDeriativeEntry next() {
			itIdx--;
			return new SparseDeriativeEntry(indices[itIdx], derivatives[itIdx]);
		}

		/**
		 * {@inheritDoc }
		 */
		@Override
		public void remove() {
			fillIdx--;
			if(itIdx > fillIdx){
				itIdx--;
			}
		}

		/**
		 * {@inheritDoc }
		 */
		@Override
		public void add(int paramIdx, double derivative) {
			indices[fillIdx] = paramIdx;
			derivatives[fillIdx] = derivative;
			fillIdx++;
		}
	}

	/**
	 * This is a representation with just one entry.
	 */
	public static class TrivialSparseLocalDerivative
			implements SparseLocalDerivative, Iterator<SparseDeriativeEntry> {

		private SparseDeriativeEntry entry = null;
		private boolean hasNext = false;

		public TrivialSparseLocalDerivative() {
		}

		/**
		 * {@inheritDoc }
		 */
		@Override
		public void add(int paramIdx, double derivative) {
			entry = new SparseDeriativeEntry(paramIdx, derivative);
		}

		/**
		 * {@inheritDoc }
		 */
		@Override
		public Iterator<SparseDeriativeEntry> iterator() {
			hasNext = true;
			return this;
		}

		/**
		 * {@inheritDoc }
		 */
		@Override
		public boolean hasNext() {
			return hasNext;
		}

		/**
		 * {@inheritDoc }
		 */
		@Override
		public void remove() {
			entry = null;
		}

		/**
		 * {@inheritDoc }
		 */
		@Override
		public SparseDeriativeEntry next() {
			if(!hasNext){
				return null;
			}
			hasNext = false;
			return entry;
		}
	}

	public static class EmptySparseLocalEntry implements SparseLocalDerivative, Iterator<SparseDeriativeEntry> {

		/**
		 * {@inheritDoc }
		 */
		@Override
		public void add(int paramIdx, double derivative) {
			throw new UnsupportedOperationException("This is empty and nothing can be added!");
		}

		/**
		 * {@inheritDoc }
		 */
		@Override
		public Iterator<SparseDeriativeEntry> iterator() {
			return this;
		}

		/**
		 * {@inheritDoc }
		 */
		@Override
		public boolean hasNext() {
			return false;
		}

		/**
		 * {@inheritDoc }
		 */
		@Override
		public SparseDeriativeEntry next() {
			return null;
		}

		/**
		 * {@inheritDoc }
		 */
		@Override
		public void remove() {
			throw new UnsupportedOperationException("This is empty and nothing can be removed!");
		}

	}

	public static class SparseDeriativeEntry implements Map.Entry<Integer, Double> {

		private final int parameterIndex;
		private double derivative;

		public SparseDeriativeEntry(int parameterIndex, double derivative) {
			this.parameterIndex = parameterIndex;
			this.derivative = derivative;
		}

		@Override
		public Integer getKey() {
			return parameterIndex;
		}

		@Override
		public Double getValue() {
			return derivative;
		}

		public double getDerivative() {
			return derivative;
		}

		public int getParameterIndex() {
			return parameterIndex;
		}

		@Override
		public Double setValue(Double arg0) {
			final Double old = this.derivative;
			this.derivative = arg0;
			return old;
		}

	}
}
