/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment.sequence;

import de.unibi.citec.tcs.alignment.comparators.CharStatComparator;
import de.unibi.citec.tcs.alignment.comparators.Comparator;
import de.unibi.citec.tcs.alignment.comparators.L1NormComparator;
import de.unibi.citec.tcs.alignment.comparators.ReplacementComparator;
import de.unibi.citec.tcs.alignment.comparators.ReplacementCosts;

/**
 * This specifies a keyword via its name and the type of its values. Further
 * specifications might be required by subclasses.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public abstract class KeywordSpecification {

	private final String keyword;
	private final ValueType type;

	/**
	 * Sets up a KeywordSpecification that defines a type for the given keyword.
	 *
	 * @param keyword a keyword.
	 * @param type the correct type for that keyword.
	 */
	public KeywordSpecification(String keyword, ValueType type) {
		this.keyword = keyword;
		this.type = type;
	}

	/**
	 * Returns the keyword.
	 *
	 * @return the keyword.
	 */
	public String getKeyword() {
		return keyword;
	}

	/**
	 * Returns the specified type for that keyword.
	 *
	 * @return the specified type for that keyword.
	 */
	public ValueType getType() {
		return type;
	}

	/**
	 *
	 * @param value a value.
	 *
	 * @return true if the given value is in order according to this
	 * specification, meaning that it has the correct type.
	 */
	public boolean validate(final Value value) {
		return value.getType() == type;
	}

	/**
	 * This returns a valid Comparator for this KeywordSpecification.
	 *
	 * @return a valid Comparator for this KeywordSpecification.
	 */
	public Comparator getDefaultComparator() {
		switch (type) {
			case SYMBOLIC:
				final Alphabet alphabet = ((SymbolicKeywordSpecification) this).
					getAlphabet();
				return new ReplacementComparator(
					new ReplacementCosts(alphabet, true, true));
			case VECTOR:
				final int dimensions = ((VectorialKeywordSpecification) this).getLength();
				return new L1NormComparator(dimensions);
			case STRING:
				return new CharStatComparator();
			default:
				throw new UnsupportedOperationException("The value type " + type + " has no default!");
		}
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public int hashCode() {
		int hash = 3;
		hash = 53 * hash + (this.keyword != null ? this.keyword.hashCode() : 0);
		hash = 53 * hash + (this.type != null ? this.type.hashCode() : 0);
		return hash;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final KeywordSpecification other = (KeywordSpecification) obj;
		if ((this.keyword == null) ? (other.keyword != null) : !this.keyword.equals(other.keyword)) {
			return false;
		}
		if (this.type != other.type) {
			return false;
		}
		return true;
	}

	@Override
	public String toString() {
		return keyword + " : " + type;
	}

}
