/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment.sequence;

import java.util.Arrays;

/**
 * This is a node in a Sequence storing values for the keywords defined in the
 * sequence.
 *
 * More formally: This Node is defined according to a NodeSpecification that
 * defines a set of keywords K as well as for each keyword k in K the allowed
 * space of values V_k. Then this node is a mapping from keywords to concrete
 * values:
 *
 * n(k) = v_k where v_k in V_k
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class Node {

    private final Sequence sequence;
    private final Value[] valueMap;

    public Node(Sequence sequence) {
        this.sequence = sequence;
        valueMap = new Value[sequence.getNodeSpecificiation().size()];
    }

    /**
     * Returns the sequence this node belongs to.
     *
     * @return the sequence this node belongs to.
     */
    public Sequence getSequence() {
        return sequence;
    }

 
    /**
     * Returns the value stored in this Node for the given keyword or null of no
     * value is stored for the given keyword in this node.
     *
     * @param keyword a keyword
     * @return the value stored in this Node for the given keyword or null of no
     * value is stored for the given keyword in this node.
     *
     * @throws UnsupportedOperationException is thrown if the given keyword is
     * not supported by the NodeSpecification.
     */
    public Value getValue(final String keyword) throws UnsupportedOperationException {
        if (!sequence.getNodeSpecificiation().hasKeyword(keyword)) {
            throw new UnsupportedOperationException("The given keyword is not defined in the sequence!");
        }
        return valueMap[sequence.getNodeSpecificiation().getKeywordIndex(keyword)];
    }

    /**
     * Returns the value stored in this Node for the given keyword or null of no
     * value is stored for the given keyword in this node.
     *
     * @param index a keyword index.
     * @return the value stored in this Node for the given keyword or null of no
     * value is stored for the given keyword in this node.
     *
     * @throws ArrayIndexOutOfBoundsException is thrown if the given index is
     * too large or too small.
     */
    public Value getValue(final int index) throws ArrayIndexOutOfBoundsException {
        return valueMap[index];
    }

    /**
     * Sets the value for the given keyword.
     *
     * @param keyword a keyword.
     * @param value a value with the correct type.
     *
     * @throws UnsupportedOperationException is thrown if the given keyword is
     * not supported by the NodeSpecification or if the given value has the
     * wrong type.
     */
    public void setValue(final String keyword, final Value value)
            throws UnsupportedOperationException {
        if (!sequence.getNodeSpecificiation().hasKeyword(keyword)) {
            throw new UnsupportedOperationException("The given keyword is not defined in the sequence!");
        }
        setValue(sequence.getNodeSpecificiation().getKeywordIndex(keyword), value);
    }

    /**
     * Sets the value for the given keyword.
     *
     * @param index a keyword index.
     * @param value a value with the correct type.
     *
     * @throws UnsupportedOperationException is thrown if the given value has
     * the wrong type.
     * @throws ArrayIndexOutOfBoundsException is thrown if the given index is
     * too large or too small.
     */
    public void setValue(final int index, final Value value)
            throws UnsupportedOperationException, ArrayIndexOutOfBoundsException {
        if (!sequence.getNodeSpecificiation().getKeywordSpecification(index).validate(value)) {
            throw new UnsupportedOperationException("The given value " + value
                    + " does not meet the given specification for the keyword!");
        }
        valueMap[index] = value;
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public int hashCode() {
        int hash = 3;
        hash = 59 * hash + Arrays.deepHashCode(this.valueMap);
        return hash;
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final Node other = (Node) obj;
        if (!Arrays.deepEquals(this.valueMap, other.valueMap)) {
            return false;
        }
        return true;
    }
}
