/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment.sequence;

import java.util.Arrays;

/**
 * This specifies the contents of the nodes in a given sequence. In essence this
 * is just a collection of keyword specifications, defining all possible
 * keywords that may be used in nodes as well as the corresponding types.
 *
 * More formally speaking a NodeSpecification defines a set of keywords K as
 * well as the value spaces V_k for each keyword k in K.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class NodeSpecification extends IndexingScheme {

	private final KeywordSpecification[] keywordSpecs;

	public NodeSpecification(final KeywordSpecification[] keywordSpecs) {
		super(extractKeywords(keywordSpecs));
		this.keywordSpecs = keywordSpecs;
	}

	private static String[] extractKeywords(KeywordSpecification[] keywordSpecs) {
		final String[] out = new String[keywordSpecs.length];
		for (int i = 0; i < out.length; i++) {
			out[i] = keywordSpecs[i].getKeyword();
		}
		return out;
	}

	/**
	 * Returns the specification for the given keyword.
	 *
	 * @param keyword a keyword.
	 *
	 * @return the specification for the given keyword.
	 */
	public KeywordSpecification getKeywordSpecification(final String keyword) {
		return getKeywordSpecification(getKeywordIndex(keyword));
	}

	/**
	 * Returns the specification for the given keyword.
	 *
	 * @param index a keyword index.
	 *
	 * @return the specification for the given keyword.
	 */
	public KeywordSpecification getKeywordSpecification(final int index) {
		return keywordSpecs[index];
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public int hashCode() {
		int hash = 5;
		hash = 71 * hash + super.hashCode();
		hash = 71 * hash + Arrays.deepHashCode(this.keywordSpecs);
		return hash;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final NodeSpecification other = (NodeSpecification) obj;
		if (!super.equals(obj)) {
			return false;
		}
		if (!Arrays.deepEquals(this.keywordSpecs, other.keywordSpecs)) {
			return false;
		}
		return true;
	}

	@Override
	public String toString() {
		if (size() == 0) {
			return "<empty NodeSpecification>";
		}
		final StringBuilder builder = new StringBuilder();
		for (int k = 0; k < size(); k++) {
			builder.append(keywordSpecs[k].toString());
			builder.append("\n");
		}
		builder.delete(builder.length() - 1, builder.length());
		return builder.toString();
	}

}
