/*
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment.wrappers;

import de.unibi.citec.tcs.alignment.AlignmentSpecification;
import de.unibi.citec.tcs.alignment.sequence.Alphabet;
import de.unibi.citec.tcs.alignment.sequence.KeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.Node;
import de.unibi.citec.tcs.alignment.sequence.NodeSpecification;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.sequence.StringKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.StringValue;
import de.unibi.citec.tcs.alignment.sequence.SymbolicKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.SymbolicValue;
import de.unibi.citec.tcs.alignment.sequence.VectorialKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.VectorialValue;
import java.util.Random;

/**
 * This class is more for demonstration purposes and creates random
 * sequences according to given input parameters.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class RandomSequenceGenerator {

	public static final Alphabet DEFAULTALPHABET = new Alphabet("a|b|c|d|e");
	public static final int DEFAULTVECTORIALLENGTH = 3;
	public static final int DEFAULTSTRINGLENGTH = 30;

	/**
	 * This generates an AlignmentSpecification and implicitly a
	 * NodeSpecification for K keywords. For the first keyword symbolic contents
	 * are generated, for the second keyword vectorial contents and for the
	 * third keyword strings. This scheme is iterated again and again.
	 *
	 * @param K the number of keywords
	 * @return an AlignmentSpecification, also containing a NodeSpecification.
	 */
	public static AlignmentSpecification generateSpecification(int K) {
		final KeywordSpecification[] keywordSpecs = new KeywordSpecification[K];
		for (int k = 0; k < K; k++) {
			final String keyword = "keyword" + k;
			switch (k % 3) {
				case 0:
					keywordSpecs[k] = new SymbolicKeywordSpecification(
							DEFAULTALPHABET, keyword);
					break;
				case 1:
					keywordSpecs[k] = new VectorialKeywordSpecification(
							DEFAULTVECTORIALLENGTH, keyword);
					break;
				case 2:
					keywordSpecs[k] = new StringKeywordSpecification(keyword);
					break;
				default:
					throw new UnsupportedOperationException("Impossible error: Modulo 3 was not between 0-2");
			}
		}
		final NodeSpecification nodeSpec = new NodeSpecification(keywordSpecs);
		return new AlignmentSpecification(nodeSpec);
	}

	/**
	 * This generates N random sequences of length L each with K keywords.
	 * For the first keyword symbolic contents are generated, for the second
	 * keyword vectorial contents and for the third keyword strings. This scheme
	 * is iterated again and again.
	 *
	 * @param N the number of sequences to be generated.
	 * @param L the length of each sequence.
	 * @param K the number of keywords on each sequence.
	 * @return the generates sequences.
	 */
	public static Sequence[] generateSequences(int N, int L, int K) {
		final KeywordSpecification[] keywordSpecs = new KeywordSpecification[K];
		for (int k = 0; k < K; k++) {
			final String keyword = "keyword" + k;
			switch (k % 3) {
				case 0:
					keywordSpecs[k] = new SymbolicKeywordSpecification(
							DEFAULTALPHABET, keyword);
					break;
				case 1:
					keywordSpecs[k] = new VectorialKeywordSpecification(
							DEFAULTVECTORIALLENGTH, keyword);
					break;
				case 2:
					keywordSpecs[k] = new StringKeywordSpecification(keyword);
					break;
				default:
					throw new UnsupportedOperationException("Impossible error: Modulo 3 was not between 0-2");
			}
		}
		return generateSequences(N, L, new NodeSpecification(keywordSpecs));
	}

	/**
	 * This generates N random sequences of length L each according to
	 * the given NodeSpecification.
	 *
	 * @param N the number of sequences to be generated.
	 * @param L the length of each sequence.
	 * @param nodeSpec the node specification for all sequences.
	 * @return the generates sequences.
	 */
	public static Sequence[] generateSequences(int N, int L,
			NodeSpecification nodeSpec) {
		final int K = nodeSpec.size();
		final Sequence[] out = new Sequence[N];
		final Random rand = new Random();
		for (int n = 0; n < N; n++) {
			out[n] = new Sequence(nodeSpec);
			for (int l = 0; l < L; l++) {
				final Node node = new Node(out[n]);
				out[n].getNodes().add(node);
				for (int k = 0; k < K; k++) {
					final KeywordSpecification keywordSpec
							= nodeSpec.getKeywordSpecification(k);
					switch (keywordSpec.getType()) {
						case SYMBOLIC: {
							//add random symbol.
							final Alphabet alpha
									= ((SymbolicKeywordSpecification) keywordSpec).getAlphabet();
							final int r = rand.nextInt(alpha.size());
							final SymbolicValue randomSymbol = new SymbolicValue(
									alpha, alpha.getSymbol(r));
							node.setValue(k, randomSymbol);
						}
						break;
						case VECTOR: {
							//add random number.
							final int D = ((VectorialKeywordSpecification) keywordSpec).getLength();
							final double[] content = new double[D];
							for (int d = 0; d < D; d++) {
								content[d] = rand.nextGaussian();
							}
							node.setValue(k, new VectorialValue(content));
						}
						break;
						case STRING: {
							//add random string.
							final StringBuilder builder = new StringBuilder();
							for (int i = 0; i < DEFAULTSTRINGLENGTH; i++) {
								char c = (char) (97 + rand.nextInt(26));
								builder.append(c);
							}
							node.setValue(k, new StringValue(builder.toString()));
						}
						break;
						default:
							throw new UnsupportedOperationException(
									"Unsupported ValueType! " + keywordSpec.getType());
					}
				}
			}
		}
		return out;
	}

}
