/*
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment.wrappers;

import de.unibi.citec.tcs.alignment.AlignmentPath;
import de.unibi.citec.tcs.alignment.AlignmentSpecification;
import de.unibi.citec.tcs.alignment.Operation;
import de.unibi.citec.tcs.alignment.comparators.Comparator;
import de.unibi.citec.tcs.alignment.comparators.GapComparator;
import de.unibi.citec.tcs.alignment.comparators.L1NormComparator;
import de.unibi.citec.tcs.alignment.sequence.KeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.Node;
import de.unibi.citec.tcs.alignment.sequence.NodeSpecification;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.sequence.VectorialKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.VectorialValue;

/**
 * This provides an easier interface for simple vectorial sequences.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class VectorialSequences {

	public static final String STDKEYWORD = "content";

	/**
	 * Transforms the given arrays to sequences. It is assumed that each
	 * entry arrays[s] is a sequence.
	 * Your sequences are allowed to have different lengths.
	 *
	 * @param arrays an array of double arrays that shall be converted to
	 * sequences.
	 * @return an array of Sequences.
	 */
	public static Sequence[] toSequences(double[][] arrays) {
		final KeywordSpecification[] keywordSpecs = {new VectorialKeywordSpecification(1, STDKEYWORD)};
		final NodeSpecification nodeSpec = new NodeSpecification(keywordSpecs);
		final Sequence[] sequences = new Sequence[arrays.length];
		for (int s = 0; s < arrays.length; s++) {
			sequences[s] = new Sequence(nodeSpec);
			for (int i = 0; i < arrays[s].length; i++) {
				final Node node = new Node(sequences[s]);
				node.setValue(0, new VectorialValue(new double[]{arrays[s][i]}));
				sequences[s].getNodes().add(node);
			}
		}
		return sequences;
	}

	/**
	 * Transforms the given arrays to sequences. It is assumed that each
	 * entry arrays[s] is a sequence, each entry arrays[s][i] is a node and
	 * that the last dimension is consistent for all nodes. This method will
	 * interpret the last dimension as <em>one</em> property.
	 * If you want to label each entry in the last dimension differently please
	 * use toSequences(double[][][] arrays, String[] keywords).
	 * Your sequences are allowed to have different lengths.
	 *
	 * @param arrays an array of double matrices that shall be converted to
	 * sequences.
	 * @return an array of Sequences.
	 */
	public static Sequence[] toSequences(double[][][] arrays) {
		//first check if the last dimension is consistent.
		int N = 0;
		for (int s = 0; s < arrays.length; s++) {
			for (int i = 0; i < arrays[s].length; i++) {
				if (arrays[s][i] == null || arrays[s][i].length == 0) {
					throw new UnsupportedOperationException("Node " + i + " of sequence " + s + " is empty!");
				}
				if (N == 0) {
					N = arrays[s][i].length;
				} else if (N != arrays[s][i].length) {
					throw new UnsupportedOperationException(
							"Length of node " + i + " in sequence " + s
							+ " was not consistent with the rest of the data set!");
				}
			}
		}
		//then construct the sequence set.
		final KeywordSpecification[] keywordSpecs = {new VectorialKeywordSpecification(N, STDKEYWORD)};
		final NodeSpecification nodeSpec = new NodeSpecification(keywordSpecs);
		final Sequence[] sequences = new Sequence[arrays.length];
		for (int s = 0; s < arrays.length; s++) {
			sequences[s] = new Sequence(nodeSpec);
			for (int i = 0; i < arrays[s].length; i++) {
				final Node node = new Node(sequences[s]);
				node.setValue(0, new VectorialValue(arrays[s][i]));
				sequences[s].getNodes().add(node);
			}
		}
		return sequences;
	}

	/**
	 * Transforms the given arrays to sequences. It is assumed that each
	 * entry arrays[s] is a sequence, each entry arrays[s][i] is a node and
	 * that the last dimension is consistent for all nodes. This method will
	 * interpret the last dimension as <em>different</em> properties.
	 * Your sequences are allowed to have different lengths.
	 *
	 * @param arrays an array of double matrices that shall be converted to
	 * sequences.
	 * @param keywords an array containing a name for each dimension of the
	 * input sequences.
	 * @return an array of Sequences.
	 */
	public static Sequence[] toSequences(double[][][] arrays, String[] keywords) {
		//first check if the last dimension is consistent.
		final int K = keywords.length;
		for (int s = 0; s < arrays.length; s++) {
			for (int i = 0; i < arrays[s].length; i++) {
				if (arrays[s][i] == null || arrays[s][i].length == 0) {
					throw new UnsupportedOperationException("Node " + i + " of sequence " + s + " is empty!");
				}
				if (K != arrays[s][i].length) {
					throw new UnsupportedOperationException(
							"Length of node " + i + " in sequence " + s
							+ " was not consistent with the length of the keyword array!");
				}
			}
		}
		//construct the keyword specifications.
		final KeywordSpecification[] keywordSpecs = new KeywordSpecification[K];
		for (int k = 0; k < K; k++) {
			keywordSpecs[k] = new VectorialKeywordSpecification(1, keywords[k]);
		}
		final NodeSpecification nodeSpec = new NodeSpecification(keywordSpecs);
		//then construct the sequences.
		final Sequence[] sequences = new Sequence[arrays.length];
		for (int s = 0; s < arrays.length; s++) {
			sequences[s] = new Sequence(nodeSpec);
			for (int i = 0; i < arrays[s].length; i++) {
				final Node node = new Node(sequences[s]);
				for (int k = 0; k < K; k++) {
					node.setValue(k, new VectorialValue(new double[]{arrays[s][i][k]}));
				}
				sequences[s].getNodes().add(node);
			}
		}
		return sequences;
	}

	/**
	 * Transforms the given arrays to sequences. It is assumed that each
	 * entry arrays[s] is a sequence, each entry arrays[s][i] is a node and
	 * that the last dimension is consistent for all nodes. This method will
	 * interpret the last dimension as <em>different</em> properties. You
	 * can set the distribution of dimensions to properties manually.
	 * Your sequences are allowed to have different lengths.
	 *
	 * @param arrays an array of double matrices that shall be converted to
	 * sequences.
	 * @param keywords an array containing a name for each property of the
	 * input resulting sequence.
	 * @param propertyDist the distribution of sequence dimensions to properties
	 * in terms of an index set. Say you want to have three properties, labelled
	 * "pos", "rot", "alpha", where "pos" and "rot" should be three-dimensional
	 * vectors each, defining the cartesian coordinates of an object in 3d-space
	 * and the rotation of it in terms of axis angles. The last property is a
	 * scalar specifying the transparency of the object. Then your input arrays
	 * should have length 7 in the third dimension, your keywords array should
	 * be
	 * {"pos","rot","alpha"} and your propertyDist array should be [3,6],
	 * meaning that of those 7 entries the entries 0 to 2 should be mapped to
	 * the "pos" property, the entries 3 to 5 should be mapped to the "rot"
	 * property and the last entry (6) should be mapped to "alpha".
	 * @return an array of Sequences.
	 */
	public static Sequence[] toSequences(double[][][] arrays, String[] keywords,
			int[] propertyDist) {
		//first check if the last dimension is consistent.
		int N = 0;
		for (int s = 0; s < arrays.length; s++) {
			for (int i = 0; i < arrays[s].length; i++) {
				if (arrays[s][i] == null || arrays[s][i].length == 0) {
					throw new UnsupportedOperationException("Node " + i + " of sequence " + s + " is empty!");
				}
				if (N == 0) {
					N = arrays[s][i].length;
				} else if (N != arrays[s][i].length) {
					throw new UnsupportedOperationException(
							"Length of node " + i + " in sequence " + s
							+ " was not consistent with the rest of the data set!");
				}
			}
		}
		//check if the propertyDist array is properly defined.
		final int K = keywords.length;
		if (propertyDist.length != K - 1) {
			throw new UnsupportedOperationException(K + " different keywords "
					+ "given but the propertyDist array had "
					+ propertyDist.length + " entries! We expected " + (K - 1) + " entries!");
		}

		for (int k = 0; k < K - 1; k++) {
			if (propertyDist[k] <= 0) {
				throw new UnsupportedOperationException("The entries of the "
						+ "propertyDist array must be larger zero!");
			}
			if (k > 0) {
				if (propertyDist[k] < propertyDist[k - 1]) {
					throw new UnsupportedOperationException("The entries of the "
							+ "propertyDist array must increasing!");
				}
			}
			if (propertyDist[k] > N - 1) {
				throw new UnsupportedOperationException("The entries of the "
						+ "propertyDist array must be smaller than the number "
						+ "of dimensions, which is " + N + " in this case!");
			}
		}
		//construct the keyword specifications.
		final KeywordSpecification[] keywordSpecs = new KeywordSpecification[K];
		for (int k = 0; k < K; k++) {
			final int startIdx;
			if (k == 0) {
				startIdx = 0;
			} else {
				startIdx = propertyDist[k - 1];
			}
			final int endIdx;
			if (k == K - 1) {
				endIdx = N - 1;
			} else {
				endIdx = propertyDist[k] - 1;
			}
			keywordSpecs[k] = new VectorialKeywordSpecification(
					endIdx - startIdx + 1, keywords[k]);
		}
		final NodeSpecification nodeSpec = new NodeSpecification(keywordSpecs);
		//then construct the sequences.
		final Sequence[] sequences = new Sequence[arrays.length];
		for (int s = 0; s < arrays.length; s++) {
			sequences[s] = new Sequence(nodeSpec);
			for (int i = 0; i < arrays[s].length; i++) {
				final Node node = new Node(sequences[s]);
				for (int k = 0; k < K; k++) {
					final int startIdx;
					if (k == 0) {
						startIdx = 0;
					} else {
						startIdx = propertyDist[k - 1];
					}
					final int endIdx;
					if (k == K - 1) {
						endIdx = N - 1;
					} else {
						endIdx = propertyDist[k] - 1;
					}
					final double[] vector = new double[endIdx - startIdx + 1];
					for (int n = startIdx; n <= endIdx; n++) {
						vector[n - startIdx] = arrays[s][i][n];
					}
					node.setValue(k, new VectorialValue(vector));
				}
				sequences[s].getNodes().add(node);
			}
		}
		return sequences;
	}

	/**
	 * Sets up an AlignmentSpecification for an array of vectorial sequences.
	 * This uses a default comparator (L1 norm and hyperbolic normalization).
	 *
	 * @param dataSpace an array of sequences.
	 * @return an AlignmentSpecification.
	 */
	public static AlignmentSpecification setUpSpecification(Sequence[] dataSpace) {
		if (dataSpace == null || dataSpace.length == 0) {
			throw new UnsupportedOperationException("The given data space is empty!");
		}
		//construct an array of L1NormComparators.
		//check for NodeSpecification consistency.
		final NodeSpecification nodeSpec = dataSpace[0].getNodeSpecificiation();
		for (int s = 1; s < dataSpace.length; s++) {
			if (!nodeSpec.equals(dataSpace[s].getNodeSpecificiation())) {
				throw new UnsupportedOperationException("The NodeSpecifications in the given data space are not consistent!");
			}
		}

		final GapComparator[] comparators = new GapComparator[nodeSpec.size()];
		for (int k = 0; k < comparators.length; k++) {
			final VectorialKeywordSpecification keySpec
					= (VectorialKeywordSpecification) nodeSpec.getKeywordSpecification(k);
			comparators[k] = new L1NormComparator(keySpec.getLength());
		}
		return setUpSpecification(dataSpace, comparators);
	}

	/**
	 * Sets up an AlignmentSpecification for an array of vectorial sequences.
	 * This uses the given comparator.
	 *
	 * @param dataSpace an array of sequences.
	 * @param comparator a comparator that works on VectorialValues.
	 * @return an AlignmentSpecification.
	 */
	public static AlignmentSpecification setUpSpecification(Sequence[] dataSpace,
			GapComparator<VectorialValue> comparator) {
		if (dataSpace == null || dataSpace.length == 0) {
			throw new UnsupportedOperationException("The given data space is empty!");
		}
		//check for NodeSpecification consistency.
		final NodeSpecification nodeSpec = dataSpace[0].getNodeSpecificiation();
		for (int s = 1; s < dataSpace.length; s++) {
			if (!nodeSpec.equals(dataSpace[s].getNodeSpecificiation())) {
				throw new UnsupportedOperationException("The NodeSpecifications in the given data space are not consistent!");
			}
		}
		final String[] keywords = nodeSpec.getKeywords();
		final Comparator[] comparators = new Comparator[keywords.length];
		for (int k = 0; k < keywords.length; k++) {
			comparators[k] = comparator;
		}
		return new AlignmentSpecification(nodeSpec, keywords, comparators);
	}

	/**
	 * Sets up an AlignmentSpecification for an array of vectorial sequences.
	 * This uses the given comparators.
	 *
	 * @param dataSpace an array of sequences.
	 * @param comparators a comparator for each dimension that works on
	 * VectorialValues.
	 * @return an AlignmentSpecification.
	 */
	public static AlignmentSpecification setUpSpecification(Sequence[] dataSpace,
			GapComparator[] comparators) {
		if (dataSpace == null || dataSpace.length == 0) {
			throw new UnsupportedOperationException("The given data space is empty!");
		}
		//check for NodeSpecification consistency.
		final NodeSpecification nodeSpec = dataSpace[0].getNodeSpecificiation();
		for (int s = 1; s < dataSpace.length; s++) {
			if (!nodeSpec.equals(dataSpace[s].getNodeSpecificiation())) {
				throw new UnsupportedOperationException("The NodeSpecifications in the given data space are not consistent!");
			}
		}
		final String[] keywords = nodeSpec.getKeywords();
		return new AlignmentSpecification(nodeSpec, keywords, comparators);
	}

	/**
	 * Sets up an AlignmentSpecification for an array of vectorial sequences.
	 * This uses a default comparator (L1 norm and hyperbolic normalization).
	 *
	 * @param dataSpace an array of sequences.
	 * @param weights a weighting putting emphasis on certain properties.
	 * @return an AlignmentSpecification.
	 */
	public static AlignmentSpecification setUpSpecification(Sequence[] dataSpace,
			double[] weights) {
		if (dataSpace == null || dataSpace.length == 0) {
			throw new UnsupportedOperationException("The given data space is empty!");
		}
		//construct an array of L1NormComparators.
		//check for NodeSpecification consistency.
		final NodeSpecification nodeSpec = dataSpace[0].getNodeSpecificiation();
		for (int s = 1; s < dataSpace.length; s++) {
			if (!nodeSpec.equals(dataSpace[s].getNodeSpecificiation())) {
				throw new UnsupportedOperationException("The NodeSpecifications in the given data space are not consistent!");
			}
		}
		final GapComparator[] comparators = new GapComparator[nodeSpec.size()];
		for (int k = 0; k < comparators.length; k++) {
			final VectorialKeywordSpecification keySpec
					= (VectorialKeywordSpecification) nodeSpec.getKeywordSpecification(k);
			comparators[k] = new L1NormComparator(keySpec.getLength());
		}
		return setUpSpecification(dataSpace, comparators, weights);
	}

	/**
	 * Sets up an AlignmentSpecification for an array of vectorial sequences.
	 * This uses the given comparator.
	 *
	 * @param dataSpace an array of sequences.
	 * @param comparator a comparator that works on VectorialValues.
	 * @param weights a weighting putting emphasis on certain properties.
	 * @return an AlignmentSpecification.
	 */
	public static AlignmentSpecification setUpSpecification(Sequence[] dataSpace,
			GapComparator<VectorialValue> comparator, double[] weights) {
		if (dataSpace == null || dataSpace.length == 0) {
			throw new UnsupportedOperationException("The given data space is empty!");
		}
		//check for NodeSpecification consistency.
		final NodeSpecification nodeSpec = dataSpace[0].getNodeSpecificiation();
		for (int s = 1; s < dataSpace.length; s++) {
			if (!nodeSpec.equals(dataSpace[s].getNodeSpecificiation())) {
				throw new UnsupportedOperationException("The NodeSpecifications in the given data space are not consistent!");
			}
		}
		final String[] keywords = nodeSpec.getKeywords();
		final Comparator[] comparators = new Comparator[keywords.length];
		for (int k = 0; k < keywords.length; k++) {
			comparators[k] = comparator;
		}
		return new AlignmentSpecification(nodeSpec, keywords, comparators, weights);
	}

	/**
	 * Sets up an AlignmentSpecification for an array of vectorial sequences.
	 * This uses the given comparator.
	 *
	 * @param dataSpace an array of sequences.
	 * @param comparators an array of comparators that works on VectorialValues.
	 * @param weights a weighting putting emphasis on certain properties.
	 * @return an AlignmentSpecification.
	 */
	public static AlignmentSpecification setUpSpecification(Sequence[] dataSpace,
			GapComparator[] comparators, double[] weights) {
		if (dataSpace == null || dataSpace.length == 0) {
			throw new UnsupportedOperationException("The given data space is empty!");
		}
		//check for NodeSpecification consistency.
		final NodeSpecification nodeSpec = dataSpace[0].getNodeSpecificiation();
		for (int s = 1; s < dataSpace.length; s++) {
			if (!nodeSpec.equals(dataSpace[s].getNodeSpecificiation())) {
				throw new UnsupportedOperationException("The NodeSpecifications in the given data space are not consistent!");
			}
		}
		final String[] keywords = nodeSpec.getKeywords();
		return new AlignmentSpecification(nodeSpec, keywords, comparators, weights);
	}

	/**
	 * Returns the sequence extensions modelled by this AlignmentPath.
	 * Gaps are modelled by null entries instead of the third dimension.
	 *
	 * @param path an AlignmentPath.
	 * @return the sequence extensions modelled by this AlignmentPath as a
	 * 2 x PathLength x Dimensions matrix.
	 */
	public static double[][][] getExtensions(AlignmentPath path) {
		int D = 0;
		for (int k = 0; k < path.getSpecificaton().size(); k++) {
			final int origK = path.getSpecificaton().getOriginalIndex(k);
			final KeywordSpecification keywordSpec
					= path.getSpecificaton().getNodeSpecification().
					getKeywordSpecification(origK);
			if (!(keywordSpec instanceof VectorialKeywordSpecification)) {
				throw new UnsupportedOperationException("The specification for keyword "
						+ path.getSpecificaton().getKeyword(k) + " is not vectorial!");
			}
			D += ((VectorialKeywordSpecification) keywordSpec).getLength();
		}

		final double[][][] out = new double[2][path.getOperations().size()][D];

		for (int i = 0; i < path.getOperations().size(); i++) {
			final Operation op = path.getOperations().get(i);
			switch (op.getType()) {
				case DELETION:
				case SKIPDELETION:
					out[1][i] = null;
					break;

			}

			int d = 0;
			for (int k = 0; k < path.getSpecificaton().size(); k++) {
				int local_d = 0;
				switch (op.getType()) {
					case DELETION:
					case SKIPDELETION:
						out[1][i] = null;
					case REPLACEMENT:
						final VectorialValue leftVal = (VectorialValue) op.getLeft().
								getValue(path.getSpecificaton().getOriginalIndex(k));
						local_d = leftVal.getVector().length;
						System.arraycopy(leftVal.getVector(), 0,
								out[0][i], d, local_d);
						break;
				}
				switch (op.getType()) {
					case INSERTION:
					case SKIPINSERTION:
						out[0][i] = null;
					case REPLACEMENT:
						final VectorialValue rightVal = (VectorialValue) op.getRight().
								getValue(path.getSpecificaton().getOriginalIndex(k));
						local_d = rightVal.getVector().length;
						System.arraycopy(rightVal.getVector(), 0,
								out[1][i], d, local_d);
						break;
				}
				d += local_d;
			}
		}

		return out;
	}

	/**
	 * Returns the difference at each operation of the given AlignmentPath.
	 *
	 * @param path an AlignmentPath.
	 * @return the difference at each operation of the given AlignmentPath as an
	 * PathLength x Dimensions-Matrix.
	 */
	public static double[][] getDifference(AlignmentPath path) {

		final double[][] out = new double[path.getOperations().size()][];

		for (int i = 0; i < path.getOperations().size(); i++) {
			final Operation op = path.getOperations().get(i);
			out[i] = op.getComparatorDistances();
		}

		return out;
	}
}
