/*
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.comparators.CharStatComparator;
import de.unibi.citec.tcs.alignment.comparators.Comparator;
import de.unibi.citec.tcs.alignment.comparators.EuclideanComparator;
import de.unibi.citec.tcs.alignment.comparators.TrivialEditComparator;
import de.unibi.citec.tcs.alignment.sequence.Alphabet;
import de.unibi.citec.tcs.alignment.sequence.KeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.Node;
import de.unibi.citec.tcs.alignment.sequence.NodeSpecification;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.sequence.StringKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.StringValue;
import de.unibi.citec.tcs.alignment.sequence.SymbolicKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.SymbolicValue;
import de.unibi.citec.tcs.alignment.sequence.VectorialKeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.VectorialValue;
import de.unibi.citec.tcs.alignment.wrappers.StringEditDistance;
import java.util.EnumMap;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;

/**
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class AbstractLocalAlignmentAlgorithmTest {

	public AbstractLocalAlignmentAlgorithmTest() {
	}

	@BeforeClass
	public static void setUpClass() {
	}

	@AfterClass
	public static void tearDownClass() {
	}

	@Before
	public void setUp() {
	}

	@After
	public void tearDown() {
	}

	/**
	 * Test of calculateAlignment method, of class
	 * AbstractLocalAlignmentAlgorithm.
	 */
	@Test
	public void testCalculateAlignment() {
		//initialize the NodeSpecification and AlignmentSpecification.
		final Alphabet alpha = new Alphabet("a|b|c|d|e|y|z");
		final NodeSpecification nodeSpec = new NodeSpecification(
				new KeywordSpecification[]{
					new SymbolicKeywordSpecification(alpha, "content")
				});
		final TrivialEditComparator comparator = new TrivialEditComparator();
		AlignmentSpecification alignSpec = new AlignmentSpecification(nodeSpec);
		alignSpec.setComparator(0, comparator);

		//set up algorithm.
		TestImpl instance = new TestImpl(alignSpec);

		//first check the trivial case of an empty sequence alignment.
		final Sequence empty = new Sequence("", alpha);
		assertEquals(0, instance.calculateAlignment(empty, empty), 0);
		//then check two other trivial cases: skipping the entire first and
		//the entire second sequence.
		final Sequence left = new Sequence("cazzzzad", alpha);
		final Sequence right = new Sequence("baae", alpha);

		assertEquals(left.getNodes().size() * comparator.getSkipDeletionCost(),
				instance.calculateAlignment(left, empty), 0);
		assertEquals(right.getNodes().size() * comparator.getSkipInsertionCost(),
				instance.calculateAlignment(empty, right), 0);

		//check the easy case of an alignment of a sequence with itself with
		//no middle skips being possible due to small length.
		assertEquals(right.getNodes().size() * comparator.getMatchCost(),
				instance.calculateAlignment(right, right), 0);
		//another easy case of an alignment with itself where middle skips
		//would in principle be possible.
		assertEquals(left.getNodes().size() * comparator.getMatchCost(),
				instance.calculateAlignment(left, left), 0);

		comparator.setSkipDeletionCost(0.2);
		comparator.setSkipInsertionCost(0.2);
		//finally a non-trivial alignment, where skips are necessary.
		double expected = 0;
		expected += 6 * comparator.getSkipDeletionCost();
		expected += 2 * comparator.getSkipInsertionCost();
		expected += 2 * comparator.getMatchCost();
		assertEquals(expected, instance.calculateAlignment(left, right), 0.01);

		//use an example, where deletions and insertions have to be used.
		final Sequence left2 = new Sequence("aaaaaaaazaaaaaaaa", alpha);
		final Sequence right2 = new Sequence("aaaaaaaayyaaaaaaaaaa", alpha);
		expected = 0;
		expected += 4 * comparator.getSkipInsertionCost();
		expected += comparator.getDeletionCost();
		expected += 16 * comparator.getMatchCost();
		assertEquals(expected, instance.calculateAlignment(left2, right2), 0.01);

		/*
		 * Now we try to set the deletion cost bigger than one. That is possible
		 * without killing normalization because it is always possible to
		 * skip both sequences which will lead to an overall score between 0 and
		 * M+N
		 */
		comparator.setMatchCost(0);
		comparator.setMismatchCost(2);
		comparator.setDeletionCost(2);
		comparator.setInsertionCost(2);
		comparator.setSkipDeletionCost(1);
		comparator.setSkipInsertionCost(1);
		final Sequence[] verses = StringEditDistance.toSequences(new String[]{
			"But now the rains weep o'er his hall, with no one there to hear.",
			"Yes now the rains weep o'er his hall, and not a soul to hear."
		});
		alignSpec = StringEditDistance.setUpSpecification(verses);
		alignSpec.setComparator(0, comparator);
		instance = new TestImpl(alignSpec);
		assertEquals(30, instance.calculateAlignment(verses[0], verses[1]), 0.01);

	}

	/**
	 * Test of calculateSkipDeletionCosts and calculateSkipInsertionCosts
	 * method, of class AbstractLocalAlignmentAlgorithm.
	 */
	@Test
	public void testCalculateSkipCosts() {
		final Alphabet alpha = new Alphabet("a|b|c");
		final KeywordSpecification[] keywordSpecs = {
			new StringKeywordSpecification("key1"),
			new VectorialKeywordSpecification(2, "key2"),
			new SymbolicKeywordSpecification(alpha, "key3")
		};
		final NodeSpecification nodeSpec = new NodeSpecification(keywordSpecs);

		final CharStatComparator stringComp = new CharStatComparator();
		final EuclideanComparator vecComp = new EuclideanComparator(2);
		final TrivialEditComparator symComp = new TrivialEditComparator();

		final AlignmentSpecification alignSpec = new AlignmentSpecification(
				nodeSpec,
				new String[]{"key1", "key2", "key3"},
				new Comparator[]{stringComp, vecComp, symComp});
		final Sequence seq = new Sequence(nodeSpec);
		final Node testNode = new Node(seq);
		final StringValue stringVal = new StringValue("test");
		testNode.setValue("key1", stringVal);
		final VectorialValue vecVal = new VectorialValue(new double[]{1, 2});
		testNode.setValue("key2", vecVal);
		final SymbolicValue symVal = new SymbolicValue(alpha, "a");
		testNode.setValue("key3", symVal);

		double[] actual = AbstractLocalAlignmentAlgorithm.
				calculateSkipDeletionCosts(testNode, alignSpec);
		assertEquals(stringComp.skipDelete(stringVal), actual[0], 0);
		assertEquals(vecComp.skipDelete(vecVal), actual[1], 0);
		assertEquals(symComp.skipDelete(symVal), actual[2], 0);

		actual = AbstractLocalAlignmentAlgorithm.
				calculateSkipInsertionCosts(testNode, alignSpec);
		assertEquals(stringComp.skipInsert(stringVal), actual[0], 0);
		assertEquals(vecComp.skipInsert(vecVal), actual[1], 0);
		assertEquals(symComp.skipInsert(symVal), actual[2], 0);
	}

	public class TestImpl extends AbstractLocalAlignmentAlgorithm<Double> {

		public TestImpl(AlignmentSpecification spec) {
			super(Double.class, spec);
		}

		public double choice(double[] choices) {
			if (choices == null || choices.length == 0) {
				throw new UnsupportedOperationException("Choice input was invalid!");
			}
			double min = choices[0];
			for (int i = 1; i < choices.length; i++) {
				if (choices[i] < min) {
					min = choices[i];
				}
			}
			return min;
		}

		@Override
		public Double transformToResult(EnumMap<Recurrence, double[][]> dp_tables,
				double[][] compareMatrix,
				double[] deletionMatrix, double[] insertionMatrix,
				double[] skipDeletionMatrix, double[] skipInsertionMatrix,
				Sequence a, Sequence b) {
			return dp_tables.get(Recurrence.SKIPDEL_START)[0][0];
		}
	}

}
