/*
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;

/**
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class SoftminTest {

	public SoftminTest() {
	}

	@BeforeClass
	public static void setUpClass() {
	}

	@AfterClass
	public static void tearDownClass() {
	}

	@Before
	public void setUp() {
	}

	@After
	public void tearDown() {
	}

	/**
	 * Test of softmin method, of class Softmin.
	 */
	@Test
	public void testSoftmin() {
		//first check if softmin returns the original number of its the only input.
		final double input1 = 1;
		for (double beta = 0; beta < 100; beta++) {
			final double result = Softmin.softmin(beta, new double[]{input1});
			assertEquals(input1, result, 0.0);
		}
		//check if for two arguments the result gets close to minimum for beta -> infinity
		final double input2 = 0.5;
		double lastResult = 1;
		for (double beta = 0; beta < 20; beta++) {
			final double result = Softmin.softmin(beta, new double[]{input1, input2});
			assertTrue(result <= lastResult);
			assertTrue(result >= input2);
			lastResult = result;
		}
		assertEquals(input2, lastResult, 0.001);
		//check if softmin also works in difficult numeric terrain.
		final double result = Softmin.softmin(10, new double[]{101, 102});
		assertEquals(101, result, 0.0001);
	}

	/**
	 * Test of calculateSoftminDerivatives method, of class Softmin.
	 */
	@Test
	public void testCalculateSoftminDerivatives() {
		//for only one input variable we expect a strict 1 for softmin', regardless of beta.
		double[] variables = {4};
		double[] expected = {1};
		for (double beta = 0; beta < 15; beta++) {
			assertArrayEquals(expected,
					Softmin.calculateSoftminDerivatives(beta, variables), 0);
		}
		//if two input variables are equal, we expect 0.5 at the output regardless of beta.
		variables = new double[]{100, 100};
		expected = new double[]{0.5, 0.5};
		for (double beta = 0; beta < 15; beta++) {
			assertArrayEquals(expected,
					Softmin.calculateSoftminDerivatives(beta, variables), 0);
		}
		//if beta is 0 we expect 0.5 for two input variables regardless of their distance.
		for (double b = 0; b < 100; b++) {
			double[] actual = Softmin.calculateSoftminDerivatives(0, new double[]{0, b});
			assertArrayEquals(expected, actual, 0);
		}
		//for high beta and unequal inputs one softmin' should be 1 and the other one 0.
		variables = new double[]{0, 4};
		expected = new double[]{1, 0};
		assertArrayEquals(expected,
				Softmin.calculateSoftminDerivatives(15, variables), 0);
		//the output should be symmetric.
		variables = new double[]{4, 0};
		expected = new double[]{0, 1};
		assertArrayEquals(expected,
				Softmin.calculateSoftminDerivatives(15, variables), 0);
		//for small beta and small distance we expect some value between 0 and 1 (with some tolerance).
		variables = new double[]{3, -2, 5, 1, -1, 4, 2, 0};
		final double[] actual = Softmin.calculateSoftminDerivatives(0.5, variables);
		final double tolerance = 0.03;
		double sum = 0;
		for (int i = 0; i < variables.length; i++) {
			assertTrue(actual[i] > 0 - tolerance);
			assertTrue(actual[i] < 1 + tolerance);
			sum += actual[i];
		}
		//the sum should be around 1.
		assertEquals(1, sum, tolerance);
		//check the order of values.
		//-2 and -1
		assertTrue(actual[1] > actual[4] - tolerance);
		//-1 and 0
		assertTrue(actual[4] > actual[7] - tolerance);
		//0 and 1
		assertTrue(actual[7] > actual[3] - tolerance);
		//1 and 2
		assertTrue(actual[3] > actual[6] - tolerance);
		//2 and 3
		assertTrue(actual[6] > actual[0] - tolerance);
		//3 and 4
		assertTrue(actual[0] > actual[5] - tolerance);
		//4 and 5
		assertTrue(actual[5] > actual[2] - tolerance);
	}
}
