/*
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.comparators.TrivialEditComparator;
import de.unibi.citec.tcs.alignment.sequence.Alphabet;
import de.unibi.citec.tcs.alignment.sequence.KeywordSpecification;
import de.unibi.citec.tcs.alignment.sequence.NodeSpecification;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.sequence.SymbolicKeywordSpecification;
import de.unibi.citec.tcs.alignment.wrappers.StringEditDistance;
import java.util.Iterator;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;

/**
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class StrictLocalAlignmentFullAlgorithmTest {

	public StrictLocalAlignmentFullAlgorithmTest() {
	}

	@BeforeClass
	public static void setUpClass() {
	}

	@AfterClass
	public static void tearDownClass() {
	}

	@Before
	public void setUp() {
	}

	@After
	public void tearDown() {
	}

	/**
	 * Test of calculateAlignment method, of class
	 * AbstractLocalAlignmentAlgorithm.
	 */
	@Test
	public void testCalculateAlignment() {

		//initialize the NodeSpecification and AlignmentSpecification.
		final Alphabet alpha = new Alphabet("a|b|c|d|e|y|z");
		final NodeSpecification nodeSpec = new NodeSpecification(
				new KeywordSpecification[]{
					new SymbolicKeywordSpecification(alpha, "content")
				});
		final TrivialEditComparator comparator = new TrivialEditComparator();
		AlignmentSpecification alignSpec = new AlignmentSpecification(nodeSpec);
		alignSpec.setComparator(0, comparator);
		comparator.setMatchCost(0);

		//set up algorithm.
		StrictLocalAlignmentFullAlgorithm instance
				= new StrictLocalAlignmentFullAlgorithm(alignSpec);

		//first check the trivial case of an empty sequence alignment.
		final Sequence empty = new Sequence("", alpha);
		AlignmentPath path = instance.calculateAlignment(empty, empty);
		assertEquals(0, path.getScore(), 0);
		assertTrue(path.getOperations().isEmpty());
		//then check two other trivial cases: skipping the entire first and
		//the entire second sequence.
		final Sequence left = new Sequence("cazzzzad", alpha);
		final Sequence right = new Sequence("baae", alpha);

		path = instance.calculateAlignment(left, empty);
		assertEquals(8, path.getOperations().size());
		assertEquals(1, path.getScore(), 0);
		for (final Operation op : path.getOperations()) {
			assertEquals(OperationType.SKIPDELETION, op.getType());
			assertNotNull(op.getLeft());
			assertNull(op.getRight());
			assertEquals(1, op.getComparatorDistances().length);
			assertEquals(comparator.getSkipDeletionCost(),
					op.getComparatorDistances()[0], 0);
		}

		path = instance.calculateAlignment(empty, right);
		assertEquals(4, path.getOperations().size());
		assertEquals(1, path.getScore(), 0);
		for (final Operation op : path.getOperations()) {
			assertEquals(OperationType.SKIPINSERTION, op.getType());
			assertNull(op.getLeft());
			assertNotNull(op.getRight());
			assertEquals(1, op.getComparatorDistances().length);
			assertEquals(comparator.getSkipInsertionCost(),
					op.getComparatorDistances()[0], 0);
		}

		//check the easy case of an alignment of a sequence with itself with
		//no middle skips being possible due to small length.
		path = instance.calculateAlignment(right, right);

		assertEquals(4, path.getOperations().size());
		assertEquals(0, path.getScore(), 0);
		for (final Operation op : path.getOperations()) {
			assertEquals(OperationType.REPLACEMENT, op.getType());
			assertNotNull(op.getLeft());
			assertNotNull(op.getRight());
			assertEquals(1, op.getComparatorDistances().length);
			assertEquals(0, op.getComparatorDistances()[0], 0);
		}
		//another easy case of an alignment with itself where middle skips
		//would in principle be possible.	
		path = instance.calculateAlignment(left, left);
		assertEquals(8, path.getOperations().size());
		assertEquals(0, path.getScore(), 0);
		for (final Operation op : path.getOperations()) {
			assertEquals(OperationType.REPLACEMENT, op.getType());
			assertNotNull(op.getLeft());
			assertNotNull(op.getRight());
			assertEquals(1, op.getComparatorDistances().length);
			assertEquals(0, op.getComparatorDistances()[0], 0);
		}

		comparator.setSkipDeletionCost(0.2);
		comparator.setSkipInsertionCost(0.2);
		//finally a non-trivial alignment, where skips are necessary.
		path = instance.calculateAlignment(left, right);
		Iterator<Operation> opIt = path.getOperations().iterator();

		assertTrue(opIt.hasNext());
		Operation op = opIt.next();
		assertEquals(OperationType.SKIPDELETION, op.getType());
		assertNull(op.getRight());
		assertNotNull(op.getLeft());
		assertEquals(1, op.getComparatorDistances().length);
		assertEquals(comparator.getSkipDeletionCost(), op.getComparatorDistances()[0], 0);

		assertTrue(opIt.hasNext());
		op = opIt.next();
		assertEquals(OperationType.SKIPINSERTION, op.getType());
		assertNotNull(op.getRight());
		assertNull(op.getLeft());
		assertEquals(1, op.getComparatorDistances().length);
		assertEquals(comparator.getSkipInsertionCost(), op.getComparatorDistances()[0], 0);

		assertTrue(opIt.hasNext());
		op = opIt.next();
		assertEquals(OperationType.REPLACEMENT, op.getType());
		assertNotNull(op.getRight());
		assertNotNull(op.getLeft());
		assertEquals(1, op.getComparatorDistances().length);
		assertEquals(0, op.getComparatorDistances()[0], 0);

		assertTrue(opIt.hasNext());
		op = opIt.next();
		assertEquals(OperationType.SKIPDELETION, op.getType());
		assertNull(op.getRight());
		assertNotNull(op.getLeft());
		assertEquals(1, op.getComparatorDistances().length);
		assertEquals(comparator.getSkipDeletionCost(), op.getComparatorDistances()[0], 0);
		assertTrue(opIt.hasNext());
		op = opIt.next();
		assertEquals(OperationType.SKIPDELETION, op.getType());
		assertNull(op.getRight());
		assertNotNull(op.getLeft());
		assertEquals(1, op.getComparatorDistances().length);
		assertEquals(comparator.getSkipDeletionCost(), op.getComparatorDistances()[0], 0);
		assertTrue(opIt.hasNext());
		op = opIt.next();
		assertEquals(OperationType.SKIPDELETION, op.getType());
		assertNull(op.getRight());
		assertNotNull(op.getLeft());
		assertEquals(1, op.getComparatorDistances().length);
		assertEquals(comparator.getSkipDeletionCost(), op.getComparatorDistances()[0], 0);
		assertTrue(opIt.hasNext());
		op = opIt.next();
		assertEquals(OperationType.SKIPDELETION, op.getType());
		assertNull(op.getRight());
		assertNotNull(op.getLeft());
		assertEquals(1, op.getComparatorDistances().length);
		assertEquals(comparator.getSkipDeletionCost(), op.getComparatorDistances()[0], 0);

		assertTrue(opIt.hasNext());
		op = opIt.next();
		assertEquals(OperationType.REPLACEMENT, op.getType());
		assertNotNull(op.getRight());
		assertNotNull(op.getLeft());
		assertEquals(1, op.getComparatorDistances().length);
		assertEquals(0, op.getComparatorDistances()[0], 0);

		assertTrue(opIt.hasNext());
		op = opIt.next();
		assertEquals(OperationType.SKIPDELETION, op.getType());
		assertNull(op.getRight());
		assertNotNull(op.getLeft());
		assertEquals(1, op.getComparatorDistances().length);
		assertEquals(comparator.getSkipDeletionCost(), op.getComparatorDistances()[0], 0);

		assertTrue(opIt.hasNext());
		op = opIt.next();
		assertEquals(OperationType.SKIPINSERTION, op.getType());
		assertNotNull(op.getRight());
		assertNull(op.getLeft());
		assertEquals(1, op.getComparatorDistances().length);
		assertEquals(comparator.getSkipInsertionCost(), op.getComparatorDistances()[0], 0);

		assertFalse(opIt.hasNext());

		//use an example, where deletions and insertions have to be used.
		final Sequence left2 = new Sequence("aaaaaaaazaaaaaaaa", alpha);
		final Sequence right2 = new Sequence("aaaaaaaayyaaaaaaaaaa", alpha);
		path = instance.calculateAlignment(left2, right2);

		boolean foundDeletion = false;
		for (final Operation o : path.getOperations()) {
			if (o.getType() == OperationType.DELETION) {
				if (foundDeletion) {
					fail();
				}
				foundDeletion = true;
			}
		}
		assertTrue(foundDeletion);

		/*
		 * Now we try to set the deletion cost bigger than one. That is possible
		 * without killing normalization because it is always possible to
		 * skip both sequences which will lead to an overall score between 0 and
		 * M+N
		 */
		comparator.setMatchCost(0);
		comparator.setMismatchCost(2);
		comparator.setDeletionCost(2);
		comparator.setInsertionCost(2);
		comparator.setSkipDeletionCost(1);
		comparator.setSkipInsertionCost(1);
		final Sequence[] verses = StringEditDistance.toSequences(new String[]{
			"But now the rains weep o'er his hall, with no one there to hear.",
			"Yes now the rains weep o'er his hall, and not a soul to hear."
		});
		alignSpec = StringEditDistance.setUpSpecification(verses);
		alignSpec.setComparator(0, comparator);
		instance = new StrictLocalAlignmentFullAlgorithm(alignSpec);
		path = instance.calculateAlignment(verses[0], verses[1]);
		//we only check the scores and the number of operations to some degree of certainty.
		assertEquals(0.24, path.getScore(), 0);
		assertEquals(71., (double) path.getOperations().size(), 5.);
	}
}
